__all__ = [
    "enter_assets_context",
    "add_css_link",
    "remove_css_link",
    "use_favicon",
    "add_import_map",
    "add_js_link",
    "remove_js_link",
    "add_js_code",
    "add_script_tag",
    "add_style_tag",
    "ComponentDependencyInfo",
    "add_component_dependency",
    "AssetsDeclaration",
    "PluginDependencyInfo",
    "register_plugin",
    "clear_registered_plugins",
    "get_active_assets",
    "register_component_extension",
]

from .context import (
    enter_assets_context,
    add_css_link,
    remove_css_link,
    use_favicon,
    add_import_map,
    add_js_link,
    remove_js_link,
    add_js_code,
    add_script_tag,
    add_style_tag,
    register_plugin,
    clear_registered_plugins,
    add_component_dependency,
    get_active_assets,
    register_component_extension,
)
from .base import AssetsDeclaration

from .component_dep import ComponentDependencyInfo
from .plugin import PluginDependencyInfo



from __future__ import annotations
from enum import IntEnum


class InputBindingType(IntEnum):
    Ref = 0
    EventContext = 1
    Data = 2
    JsFn = 3
    ElementRef = 4
    EventContextDataset = 5


class OutputSetType(IntEnum):
    Ref = 0
    RouterAction = 1
    ElementRefAction = 2
    JsCode = 3
    FileDownload = 4



from typing import Optional, Sequence

from instaui.internal.ui._expr.literal_expr import UILiteralExpr


class PreSetupAction:
    def __init__(
        self,
        *,
        code: str,
        inputs: Optional[Sequence] = None,
    ):
        self.type = "action"
        self.inputs = UILiteralExpr.try_parse_list(inputs or [])
        self.code = code


def normalize_pre_setup(pre_setup: list) -> list:
    if not pre_setup:
        return pre_setup
    first = pre_setup[0]
    # e.g [ref, True, False]
    if not isinstance(first, list):
        pre_setup = [pre_setup]

    # e.g [[ref, True, False], [ref2, False, True]]
    return [
        [x[0], UILiteralExpr.try_parse(x[1]), UILiteralExpr.try_parse(x[2])]
        for x in pre_setup
    ]



from __future__ import annotations
from typing import Optional
from instaui.internal.ui.element import Element


class Label(Element):
    def __init__(
        self,
        text: str,
        *,
        for_: Optional[str] = None,
    ):
        super().__init__("label")

        self.props(
            {
                "for": for_,
                "innerText": text,
            }
        )



from typing import Literal, Optional, Sequence, Union, cast, Tuple

TSystemModifiers = Literal["ctrl", "shift", "alt", "meta"]
TModifierGuards = Literal[
    "shift",
    "ctrl",
    "alt",
    "meta",
    "left",
    "right",
    "stop",
    "prevent",
    "self",
    "middle",
    "exact",
]
TCompatModifiers = Literal["esc", "space", "up", "left", "right", "down", "delete"]

TOtherModifier = Literal["passive", "capture", "once"]


TEventModifier = Union[
    TSystemModifiers, TModifierGuards, TCompatModifiers, TOtherModifier
]


def parse_event_modifiers(
    event_name: str,
    org_modifier: Optional[list[TEventModifier]] = None,
) -> Tuple[str, Optional[Sequence[TEventModifier]]]:
    """Parse event name and modifiers from both event_name string and org_modifier list.

    Args:
        event_name: Event name string, may contain modifiers separated by dots (e.g. 'click.stop')
        org_modifier: Optional list of additional modifiers

    Returns:
        Tuple of (cleaned_event_name, combined_modifiers) where:
            - cleaned_event_name: event name without modifiers
            - combined_modifiers: tuple of unique modifiers from both sources
    """
    parts = event_name.split(".")
    base_name = parts[0]
    modifiers = [m.strip() for m in parts[1:]] if len(parts) > 1 else []

    if not org_modifier and not modifiers:
        return base_name, None

    combined = set(modifiers)
    if org_modifier:
        combined.update(org_modifier)

    return base_name, cast(
        Sequence[TEventModifier], tuple(combined)
    ) if combined else None



import typing

from .bindable import mark_used
from .enums import InputBindingType
from .protocol import CanInputProtocol


class InputSilentData(
    CanInputProtocol,
):
    def __init__(self, value: typing.Any) -> None:
        """
        Wraps a value to mark it as "silent" for reactive tracking,
        so that changes to this value do not trigger dependent watchers
        or computed functions.

        Args:
            value (Union[BindingTrackerMixin, Any]): The value to wrap.
                Can be any data type or a reactive binding,
                which will be excluded from triggering reactive updates.


        Example:
        .. code-block:: python
            from instaui import ui, html

            a = ui.state("a")
            b = ui.state("b")
            result = ui.state("result")

            # Use silent data for 'b' so changes to 'b' don't trigger recomputation
            @ui.watch(inputs=[a, ui.slient(b)], outputs=[result])
            def only_a_changed(a: str, b: str):
                return f"{a}+{b}"

            # Only changes to 'a' will update the result
            html.input(a).classes("a")
            html.input(b).classes("b")
            html.paragraph(result)
        """
        mark_used(value)
        self.value = value

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref



from __future__ import annotations
from typing import Any, Literal, Union
from instaui.internal.ui.element import Element


class Heading(Element):
    def __init__(
        self,
        text: Union[str, Any],
        *,
        level: Literal[1, 2, 3, 4, 5, 6] = 1,
    ):
        super().__init__(f"h{level}")
        self.props(
            {
                "innerText": text,
            }
        )


class H1(Heading):
    def __init__(self, text: Union[str, Any]):
        super().__init__(text, level=1)


class H2(Heading):
    def __init__(self, text: Union[str, Any]):
        super().__init__(text, level=2)


class H3(Heading):
    def __init__(self, text: Union[str, Any]):
        super().__init__(text, level=3)


class H4(Heading):
    def __init__(self, text: Union[str, Any]):
        super().__init__(text, level=4)


class H5(Heading):
    def __init__(self, text: Union[str, Any]):
        super().__init__(text, level=5)


class H6(Heading):
    def __init__(self, text: Union[str, Any]):
        super().__init__(text, level=6)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder

from instaui.internal.ui._scope import Scope
from instaui.internal.ast import core as ast_core, expression
from instaui.internal.pipeline.ast_build.context import AstBuilderContext
from instaui.internal.ui.expr_watch import ExprWatch

if TYPE_CHECKING:
    pass


class ExprWatchParser:
    def __init__(
        self,
        builder_ctx: AstBuilderContext,
    ) -> None:
        self.ctx = builder_ctx
        self.expr = ExprBuilder(builder_ctx)

    def parse(self, scope: Scope) -> Optional[list[ast_core.ExprWatchCall]]:
        if not scope._expr_watchs:
            return None

        return [self._parse_watch(watch) for watch in scope._expr_watchs]

    def _parse_watch(self, watch: ExprWatch):
        args = expression.ObjectExpr.from_kwargs(
            code=expression.Literal(watch.code),
            on=self.expr.list_shallow(watch._on),
            bind=self.expr.object_shallow(watch._bindings) if watch._bindings else None,
            immediate=expression.Literal(watch._immediate)
            if watch._immediate
            else None,
            deep=expression.Literal(watch._deep) if watch._deep else None,
            once=expression.Literal(watch._once) if watch._once else None,
            flush=expression.Literal(watch._flush) if watch._flush else None,
        )

        return ast_core.ExprWatchCall(args)



from typing import Any, Optional
from instaui.internal.ui._expr.codec_expr import DataCodecExpr
from instaui.internal.ui.protocol import CanOutputProtocol, ObservableProtocol
from instaui.debug.source import get_source_span
from .app_index import new_scope_if_needed
from .variable import Variable
from .enums import InputBindingType, OutputSetType
from .path_var import PathVar
from .missing import MISSING
from .bindable import BindableMixin, VarableBindHelper
from .app_context import get_current_scope


class RefBase(
    PathVar,
    Variable,
    ObservableProtocol,
    CanOutputProtocol,
    BindableMixin,
):
    def __init__(
        self,
        *,
        value: Optional[Any] = MISSING,
        deep_compare: bool = False,
    ) -> None:
        self._source_span_ = get_source_span()
        new_scope_if_needed()
        define_scope = get_current_scope()
        self._bind_helper = VarableBindHelper(
            self,
            define_scope.register_ref,
            define_scope=define_scope,
            lazy_mark_used=[value],
        )

        self._org_value = value.value if isinstance(value, DataCodecExpr) else value
        self._value = value
        self._deep_compare = deep_compare

    def initial(self) -> Any:
        return self._org_value

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.Ref

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id



from typing import Any, Callable, List, Tuple, TypeVar, cast


# ---------- Operation Utilities ----------


def remove_none(key: str, value: Any) -> Tuple[str, Any, bool]:
    """
    Operation: remove entries with None values.
    Return (key, value, keep_entry)
    """
    return key, value, value is not None


def remove_empty_collection(key: str, value: Any) -> Tuple[str, Any, bool]:
    """
    Operation: remove entries with empty collections.
    """
    return key, value, (not isinstance(value, (list, dict))) or len(value) > 0


def repr_dict_keys(key: str, value: Any) -> Tuple[str, Any, bool]:
    return repr(key), value, True


# ---------- Core Processor ----------

_TData = TypeVar("_TData")


def process_dict(
    data: _TData,
    operations: List[Callable[[str, Any], Tuple[str, Any, bool]]],
    *,
    recursive: bool = True,
) -> _TData:
    """
    Apply multiple operations to a dictionary efficiently.
    operations: list of functions(key, value) -> (new_key, new_value, keep)
    """

    if not isinstance(data, dict):
        return data

    output = {}

    for key, value in data.items():
        original_key = key
        original_value = value

        # Recursively process nested data
        if recursive:
            if isinstance(value, dict):
                value = process_dict(value, operations, recursive=True)
            elif isinstance(value, list):
                value = _process_list(value, operations, recursive)

        # Apply operations sequentially
        keep = True
        for op in operations:
            key, value, keep = op(key, value)
            if not keep:
                break

        if keep:
            output[key] = value
        else:
            # Reset key/value so next item does not get affected accidentally
            key, value = original_key, original_value

    return cast(_TData, output)


def _process_list(
    data_list: List[Any], operations: List[Callable], recursive: bool
) -> List[Any]:
    result = []
    for item in data_list:
        if isinstance(item, dict):
            result.append(process_dict(item, operations, recursive=True))
        elif isinstance(item, list) and recursive:
            result.append(_process_list(item, operations, recursive=True))
        else:
            result.append(item)
    return result



from __future__ import annotations
from typing import (
    Any,
    TypeVar,
    cast,
)
from instaui.debug.api_boundary import user_api
from instaui.debug.source import get_source_span
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.app_index import new_scope_if_needed
from instaui.internal.ui.enums import InputBindingType, OutputSetType
from instaui.internal.ui.variable import Variable
from instaui.internal.ui.bindable import VarableBindHelper, BindableMixin
from instaui.internal.ui.path_var import PathVar
from instaui.internal.ui.protocol import CanInputProtocol, CanOutputProtocol

_T = TypeVar("_T")


class ConstData(Variable, PathVar, CanInputProtocol, CanOutputProtocol, BindableMixin):
    def __init__(self, value: Any = None) -> None:
        self._source_span_ = get_source_span()
        new_scope_if_needed()
        self.value = UILiteralExpr.try_parse(value)
        define_scope = get_current_scope()
        self._bind_helper = VarableBindHelper(
            self,
            define_scope.register_data,
            define_scope=define_scope,
            lazy_mark_used=[value],
        )

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    def _to_event_output_type(self) -> OutputSetType:
        raise TypeError("ConstData cannot be used as an output")

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Data

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id


TConstData = ConstData


@user_api
def const_data(value: _T) -> _T:
    return cast(_T, ConstData(value))



from __future__ import annotations

from typing import Any, Optional
from .bindable import mark_used


class Directive:
    def __init__(
        self,
        *,
        is_sys: bool,
        name: str,
        arg: Optional[str] = None,
        modifiers: Optional[list[Any]] = None,
        value: Optional[Any] = None,
    ) -> None:
        super().__init__()
        mark_used(value)
        self.name = name
        self._arg = arg
        self._is_sys = is_sys
        self._modifiers = modifiers
        self._value = value

    def __hash__(self) -> int:
        return hash(f"{self.name}:{self._arg}:{self._modifiers}")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Directive):
            return False
        return (
            self.name == other.name
            and self._arg == other._arg
            and self._modifiers == other._modifiers
        )



from pathlib import Path
import shutil
from instaui.internal.pipeline.asset_resolver.webview_resolver import AssetMaterializer
from instaui.internal.runtime.utils import path_utils
from instaui.systems.route_path_system import route_to_dirname


class FileSystemAssetMaterializer(AssetMaterializer):
    def ensure_dir(self, path: Path) -> None:
        path.mkdir(parents=True, exist_ok=True)

    def copy_file(self, src: Path, dst: Path) -> None:
        if not dst.exists():
            shutil.copy2(src, dst)

    def copy_dir(self, src: Path, dst: Path) -> None:
        if not dst.exists():
            shutil.copytree(src, dst)


class PageAssetManager:
    """
    Manages the generation and cleanup logic of WebView page resource files.
    """

    def __init__(self, assets_root: Path, clean_on_start: bool = True):
        self.assets_root = assets_root
        self.clean_on_start = clean_on_start
        self._assets_cleaned = False
        self._generated_pages: set[str] = set()

    def _clean_assets_root(self):
        """
        Cleans the entire assets root directory, executed only once.
        """
        if self.clean_on_start and not self._assets_cleaned:
            path_utils.reset_dir(self.assets_root)
            self._assets_cleaned = True

    def _clean_page_dir(self, route: str) -> Path:
        """
        Cleans a single page directory and returns its Path.
        """
        page_dir = self.assets_root.joinpath(route_to_dirname(route))
        if page_dir.exists():
            path_utils.reset_dir(page_dir)
        return page_dir

    def prepare_page_dir(self, route: str) -> Path:
        """
        Prepares directory for page generation:
        1. Clean the entire assets root directory before first page generation (optional)
        2. Clean the current page directory
        3. Return the page directory Path
        """
        self._clean_assets_root()
        return self._clean_page_dir(route)

    def mark_page_generated(self, route: str):
        """
        Marks the page as generated to avoid duplicate cleanup.
        """
        self._generated_pages.add(route)

    def is_page_generated(self, route: str) -> bool:
        """
        Determines whether the page has been generated.
        """
        return route in self._generated_pages



from typing import Any, Callable, Optional
from instaui.constants.runtime import RuntimeMode
from instaui.systems.dataclass_system import dataclass


@dataclass()
class RuntimeContext:
    mode: RuntimeMode
    route_path: Optional[str] = None
    debug_mode: bool = False
    request_prefix: str = ""
    meta: Optional[dict] = None
    path_params_getter: Optional[Callable[[str, Any], Any]] = None



from __future__ import annotations
from typing import Dict, List, Optional, Tuple

from instaui.protocol.module_import.import_item import PresetProtocol
from instaui.systems.dataclass_system import dataclass


@dataclass(frozen=True)
class _ImportItem:
    module: str
    name: str
    local: str  # Name in the current file


class ImportTable:
    def __init__(self) -> None:
        # Used local names for conflict detection
        self._used_local_names: Dict[str, int] = {}

        # Semantic deduplication table: (module, name) -> local name
        self._by_member: Dict[Tuple[str, str], str] = {}

        # Final output order
        self._items: List[_ImportItem] = []

    # ---------- public API ----------

    def use(
        self,
        *,
        module: str,
        name: str,
        alias: Optional[str] = None,
    ) -> str:
        key = (module, name)

        # ① Already imported: return directly
        if key in self._by_member:
            return self._by_member[key]

        # ② Calculate expected local name
        base_local = alias or name
        local = self._allocate_local_name(base_local)

        # ③ Record
        self._by_member[key] = local
        self._items.append(
            _ImportItem(
                module=module,
                name=name,
                local=local,
            )
        )
        return local

    def use_from_preset(
        self,
        preset: PresetProtocol,
    ) -> str:
        return self.use(
            module=preset.module_name,
            name=preset.member_name,
            alias=preset.member_alias,
        )

    def render(self) -> str:
        # Group import items by module while preserving the first occurrence order of modules
        module_groups: Dict[str, List[_ImportItem]] = {}
        module_order: List[str] = []

        for item in self._items:
            if item.module not in module_groups:
                module_groups[item.module] = []
                module_order.append(item.module)
            module_groups[item.module].append(item)

        # Generate combined import statements for each module
        lines: List[str] = []
        for module in module_order:
            items = module_groups[module]
            import_specifiers = []

            for item in items:
                if item.local == item.name:
                    import_specifiers.append(item.name)
                else:
                    import_specifiers.append(f"{item.name} as {item.local}")

            import_statement = (
                f"import {{ {', '.join(import_specifiers)} }} from '{module}';"
            )
            lines.append(import_statement)

        return "\n".join(lines)

    # ---------- internal ----------

    def _allocate_local_name(self, base: str) -> str:
        """
        Allocate a non-conflicting local name
        """
        if base not in self._used_local_names:
            self._used_local_names[base] = 0
            return base

        self._used_local_names[base] += 1
        return f"{base}_{self._used_local_names[base]}"



from abc import ABC, abstractmethod
from typing import Any, Callable, Mapping, Optional, Sequence, Union
from typing_extensions import TypeIs
import types

from instaui.internal.ui.variable import Variable
from instaui.internal.ui._scope import Scope
from instaui.internal.ui.app_context import get_current_scope


class BindableMixin(ABC):
    @property
    @abstractmethod
    def _used(self) -> bool: ...

    @abstractmethod
    def _mark_used(self) -> None: ...

    @abstractmethod
    def _mark_provided(self) -> None: ...


def is_bindable(obj: Any) -> TypeIs[BindableMixin]:
    return isinstance(obj, BindableMixin)


class VarableBindHelper:
    def __init__(
        self,
        owner: Any,
        scope_register_fn: Optional[Callable[[Any], None]] = None,
        *,
        maybe_provides: Optional[Sequence[Union[Variable, Any]]] = None,
        define_scope: Scope,
        on_register_fn: Optional[Callable[..., Any]] = None,
        lazy_mark_used: Optional[Sequence[Any]] = None,
    ):
        self._lazy_mark_used = lazy_mark_used or []
        self.maybe_provides = list(maybe_provides or ()) + [owner]

        def on_mark_used(owner: Any):
            if lazy_mark_used:
                for item in lazy_mark_used:
                    mark_used(
                        item()
                        if callable(item) and isinstance(item, types.FunctionType)
                        else item
                    )
            if scope_register_fn:
                scope_register_fn(owner)

            if on_register_fn:
                on_register_fn()

        self._on_mark_used = on_mark_used
        self._define_scope = define_scope
        self._owner = owner
        self._used = False

    @property
    def used(self) -> bool:
        return self._used

    def _try_injection(self):
        current_scope = get_current_scope()
        for target in self.maybe_provides:
            if isinstance(target, Variable):
                if current_scope is not self._define_scope:
                    if target._is_providable():
                        self._define_scope.provide(target)
                    current_scope.inject(target)

    def mark_used(self):
        self._try_injection()

        if self._used:
            return
        self._used = True
        if self._on_mark_used is not None:
            try:
                self._on_mark_used(self._owner)
            except Exception:
                raise

    def mark_provided(self):
        self._define_scope.provide(self._owner)


def mark_used(obj: Any, *, host_scope_id: Optional[int] = None):
    if isinstance(obj, (list, tuple)):
        for item in obj:
            mark_used(item, host_scope_id=host_scope_id)
    elif isinstance(obj, Mapping):
        for value in obj.values():
            mark_used(value, host_scope_id=host_scope_id)
    elif is_bindable(obj):
        if (
            host_scope_id is not None
            and isinstance(obj, Variable)
            and obj._define_scope_id > host_scope_id
        ):
            raise ValueError("Variable definition cannot be bound to an outer scope.")

        obj._mark_used()
    else:
        pass


def mark_provided(obj: BindableMixin):
    obj._mark_provided()



from typing import Final

PAGE_TITLE: Final = "insta-ui"
SCOPED_STYLE_GROUP_ID: Final = "insta-scoped-style"
DEFAULT_SLOT_NAME: Final = "default"
MOUNTED_EVENT_NAME: Final = "mounted"
STORAGE_REF_METHOD: Final = "storageRef"
INSTAUI_JS_MODULE_NAME: Final = "instaui"
USE_DARK_REF_METHOD: Final = "useDarkRef"
USE_LANGUAGE_METHOD: Final = "useLanguageRef"
USE_PAGE_TITLE_METHOD: Final = "usePageTitleRef"
USE_ROUTE_PARAM_METHOD: Final = "useRouteParam"
USE_QUERY_PARAM_METHOD: Final = "useQueryParam"

USER_ASSETS_DIR_NAME: Final = "assets"
USER_ASSETS_ICONS_DIR_NAME: Final = "icons"



from __future__ import annotations
from typing import TYPE_CHECKING
from instaui.internal.assets.base import AssetsDeclaration

if TYPE_CHECKING:
    from instaui.internal.ui.element import Element


class ZeroModeComponentExtensionProcessor:
    def process(self, assets: AssetsDeclaration) -> None:
        used_components = {
            record.component for record in assets.component_dependencies.records
        }

        for target_cls, extensions in assets.component_extensions.items():
            if not any(issubclass(used, target_cls) for used in used_components):
                continue

            self._materialize_extensions(target_cls, extensions, assets)

    def _materialize_extensions(
        self,
        target_cls: type[Element],
        extensions: dict[str, set[str]],
        assets: AssetsDeclaration,
    ):
        resolver = self._get_extension_resolver(target_cls)
        if resolver is None:
            return

        resolver.resolve_zero_extensions(extensions, assets)

    def _get_extension_resolver(self, target_cls):
        return getattr(target_cls, "__zero_extension_resolver__", None)



from instaui.internal.pipeline.assets import (
    AssetsPipeline,
    PluginCssProcessor,
    PluginImportMapProcessor,
    ZeroModeComponentCssProcessor,
    ZeroModeComponentImportMapProcessor,
    ZeroModeComponentExtensionProcessor,
)


ZERO_ASSETS_PIPELINE = AssetsPipeline(
    [
        PluginCssProcessor(),
        PluginImportMapProcessor(),
        ZeroModeComponentCssProcessor(),
        ZeroModeComponentImportMapProcessor(),
        ZeroModeComponentExtensionProcessor(),
    ]
)



from __future__ import annotations
from typing import TYPE_CHECKING

from instaui.internal.ast import core as ast_core
from instaui.internal.ast import expression
from instaui.internal.backend.invocation import BackendInvocation, BackendInvocationKind
from instaui.internal.pipeline.ast_build.exception import AstBuildError
from instaui.internal.pipeline.ast_build.pre_setup_parser import PreSetupParser
from instaui.internal.pipeline.ast_build.utils import ui_bindings_utils
from instaui.internal.pipeline.ast_build.context import AstBuilderContext
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder
from instaui.internal.ui.const_data import ConstData
from instaui.internal.ui.element_ref import ElementRef
from instaui.internal.ui.expr_event import ExprEvent
from instaui.internal.ui.js_fn import JsFn
from instaui.internal.ui.ref_base import RefBase
from instaui.internal.ui.custom_var import CustomVar
from instaui.internal.ui.expr_computed import ExprComputed
from instaui.internal.ui.web_computed import WebComputed
from instaui.internal.ui.web_event import WebEvent
from instaui.internal.ui.js_computed import JsComputed
from instaui.internal.ui.js_event import JsEvent


if TYPE_CHECKING:
    from instaui.internal.ui.variable import Variable


class VariableDeclBuilder:
    def __init__(self, ctx: AstBuilderContext) -> None:
        self.ctx = ctx
        self.expr_builder = ExprBuilder(ctx)
        self.pre_setup_parser = PreSetupParser(self.expr_builder)

    def build(self, variable: Variable) -> ast_core.VariableDecl:
        try:
            method = getattr(self, f"_build_{type(variable).__name__}", None)
            if method:
                return method(variable)

            if isinstance(variable, CustomVar):
                return self._build_CustomRef(variable)

            raise ValueError(f"Unsupported variable type: {type(variable)}")
        except Exception as e:
            raise AstBuildError(
                f"Failed to build {variable}: {e}",
                variable._source_span_,  # type: ignore
            )

    def _build_ConstData(self, variable: ConstData) -> ast_core.ConstDataDecl:
        sid = self._declare_var(variable)
        return ast_core.ConstDataDecl(
            variable._source_span_, sid, self.expr_builder.build(variable.value)
        )

    def _build_RefBase(self, variable: RefBase) -> ast_core.RefDecl:
        sid = self._declare_var(variable)
        args = self.expr_builder.object_shallow(
            {
                "value": variable._value,
                "deepCompare": variable._deep_compare or None,
            }
        )

        return ast_core.RefDecl(variable._source_span_, sid, args)

    def _build_WebComputed(self, variable: WebComputed) -> ast_core.BackendComputedDecl:
        sid = self._declare_var(variable)

        args = self.expr_builder.object_shallow(
            {
                "init": variable._init_value,
                "deepCompare": variable._deep_compare_on_input or None,
            }
        )

        return ast_core.BackendComputedDecl(variable._source_span_, sid, args)

    def _build_JsComputed(self, variable: JsComputed) -> ast_core.JsComputedDecl:
        sid = self._declare_var(variable)

        inputs, _, _ = ui_bindings_utils.analyze_observable_inputs(
            variable._inputs or []
        )

        args = self.expr_builder.object_shallow(
            {
                "code": variable.code,
                "inputs": self.expr_builder.list_shallow(inputs) or None,
                "asyncInit": variable._async_init_value,
                "deepEqOnInput": 1
                if variable._deep_compare_on_input is not False
                else None,
                "tool": variable._tool,
            }
        )

        return ast_core.JsComputedDecl(variable._source_span_, sid, args)

    def _build_JsEvent(self, variable: JsEvent) -> ast_core.JsEventDecl:
        sid = self._declare_var(variable)
        args = self.expr_builder.object_shallow(
            {
                "code": variable._code,
                "inputs": self.expr_builder.list_shallow(variable._inputs) or None,
                "outputs": self.expr_builder.list_shallow(variable._outputs) or None,
                "iptTypes": ui_bindings_utils.inputs_to_types(variable._inputs),
                "opTypes": ui_bindings_utils.outputs_to_types(variable._outputs),
            }
        )

        return ast_core.JsEventDecl(
            variable._source_span_,
            sid,
            args,
        )

    def _build_WebEvent(self, variable: WebEvent) -> ast_core.BackendEventDecl:
        sid = self._declare_var(variable)
        ref_id = self.ctx.binding_registry.register(
            BackendInvocation(
                kind=BackendInvocationKind.EVENT,
                fn=variable._fn,
                spec=variable._export_invocation_spec(),
                render_ctx=self.ctx.render_ctx,
            )
        )

        hkey = self.ctx.binding_registry.get_value(ref_id)

        f_type = ui_bindings_utils.py_function_type(variable._fn)
        pre_setup = self.pre_setup_parser.parse(variable._pre_setup)
        args = self.expr_builder.object_shallow(
            {
                "hKey": hkey,
                "fType": None if f_type == "sync" else f_type,
                "inputs": self.expr_builder.list_shallow(variable._inputs) or None,
                "outputs": self.expr_builder.list_shallow(variable._outputs) or None,
                "iptTypes": ui_bindings_utils.inputs_to_types(variable._inputs)
                if variable._inputs
                else None,
                "opTypes": ui_bindings_utils.outputs_to_types(variable._outputs),
                "preSetup": pre_setup,
            }
        )

        return ast_core.BackendEventDecl(variable._source_span_, sid, args)

    def _build_ElementRef(self, variable: ElementRef) -> ast_core.ElementRefDecl:
        sid = self._declare_var(variable)
        return ast_core.ElementRefDecl(variable._source_span_, sid)

    def _build_ExprComputed(self, variable: ExprComputed) -> ast_core.ExprComputedDecl:
        sid = self._declare_var(variable)

        args = self.expr_builder.object_shallow(
            {
                "code": expression.Literal(variable.code),
                "bind": self.expr_builder.object_shallow(variable._bindings)
                if variable._bindings
                else None,
            }
        )

        return ast_core.ExprComputedDecl(variable._source_span_, sid, args)  # type: ignore

    def _build_ExprEvent(self, variable: ExprEvent) -> ast_core.ExprEventDecl:
        sid = self._declare_var(variable)
        args = self.expr_builder.object_shallow(
            {
                "code": expression.Literal(variable._code),
                "bind": self.expr_builder.object_shallow(variable._bindings) or None,
            }
        )

        return ast_core.ExprEventDecl(variable._source_span_, sid, args)

    def _build_JsFn(self, variable: JsFn) -> ast_core.JsFnDecl:
        sid = self._declare_var(variable)
        code = variable.code
        if variable._execute_immediately:
            code = f"({code})()"
        return ast_core.JsFnDecl(
            variable._source_span_, sid, expression.RawFunctionExpr(code)
        )

    def _build_CustomRef(self, variable: CustomVar) -> ast_core.CustomRefDecl:
        sid = self._declare_var(variable)
        params = []
        if variable._args is not None:
            if isinstance(variable._args, tuple):
                params = [self.expr_builder.build(arg) for arg in variable._args]
            else:
                params = [self.expr_builder.build(variable._args)]

        return ast_core.CustomRefDecl(
            variable._source_span_,
            sid,
            variable._method.module_name,
            variable._method.method_name,
            params,
        )

    def _declare_var(self, variable: Variable):
        sid = self.ctx.var_mapper.get_or_create_id(variable)
        self.ctx.component_scope_manager.declare_var(sid)
        return sid



from __future__ import annotations
from contextlib import asynccontextmanager
from dataclasses import replace
from typing import TYPE_CHECKING, Any, Callable, Optional
from enum import Enum
from instaui.systems import file_path_system

if TYPE_CHECKING:
    from fastapi import FastAPI


class WebLauncher:
    def __init__(
        self,
        *,
        debug: bool = False,
        use_gzip: int | bool = True,
        prerender: bool = False,
        prerender_concurrency: int = 4,
    ):
        """
        Create a new server instance.

        Args:
            debug (bool): Whether to run in debug mode. In debug mode, it has the development hot-reload feature. Defaults to False.
            use_gzip (int|bool, optional):  Whether to use gzip compression. If an integer is provided, it will be used as the minimum response size for compression. If True, the default minimum size of 500 bytes will be used. If False, compression will not be used. Defaults to True.
            prerender (bool, optional): Whether to prerender the page. Defaults to False. Note: When debug=True, the prerender option will be ignored (no prerendering will be performed).
            prerender_concurrency (int, optional): The number of concurrent prerendering processes to use. Defaults to 4.
        """
        from instaui.internal.runtime.web.web_runtime import WebRuntime
        from instaui.internal.runtime.web.config import WebUserLaunchConfig

        self.launch_config = WebUserLaunchConfig(
            caller_folder_path=file_path_system.get_caller_path().parent,
            debug=debug,
            use_gzip=use_gzip,
            prerender=prerender,
            prerender_concurrency=prerender_concurrency,
        )

        self._startup_hooks: list[Callable[[], None]] = []
        self._runtime = WebRuntime(self.launch_config)

    def add_startup_hook(self, fn: Callable[[], None]):
        self._startup_hooks.append(fn)

    def run(
        self,
        host="0.0.0.0",
        port=8080,
        reload: bool = False,
        reload_dirs: str = ".",
        reload_includes: str = "*.py",
        reload_excludes: str = ".*, .py[cod], .sw.*, ~*",
        log_level="info",
        workers: Optional[int] = None,
        uds: Optional[str] = None,
        **kwargs: Any,
    ):
        from instaui.internal.runtime.web.options import WebUserRunOptions

        if self._startup_hooks:

            @asynccontextmanager
            async def lifespan_wrapper(app):
                for fn in self._startup_hooks:
                    fn()
                yield

            self._runtime.backend.router.lifespan_context = lifespan_wrapper

        opts = WebUserRunOptions(
            host=host,
            port=port,
            reload=reload,
            reload_dirs=reload_dirs,
            reload_includes=reload_includes,
            reload_excludes=reload_excludes,
            log_level=log_level,
            workers=workers,
            uds=uds,
            kwargs=kwargs,
        )

        self._runtime.run(opts)

    def run_with(
        self,
        app: FastAPI,
        *,
        prefix: str = "/instaui",
        tags: Optional[list[str | Enum]] = None,
        dependencies: Optional[list[Any]] = None,
        responses: Optional[dict[int | str, dict[str, Any]]] = None,
    ):
        """
        Mounts the InstaUI interface onto a FastAPI application with configurable routing.

        Args:
            app (FastAPI): The FastAPI application instance to mount InstaUI onto.
            prefix (str, optional): The URL prefix for InstaUI routes. Defaults to "/instaui".
            tags (Optional[list[str | Enum]], optional): OpenAPI tags for InstaUI endpoints.
            dependencies (Optional[list[Any]], optional): Dependencies to apply to all InstaUI routes.
            responses (Optional[dict[int | str, dict[str, Any]]], optional): Additional response definitions for InstaUI endpoints.

        Example:
        .. code-block:: python
            app = FastAPI()

            @app.get("/")
            def index():
                return {"message": "Hello World"}

            # www.example.com/instaui/ -> InstaUI interface
            @ui.page("/")
            def index_page():
                ui.text("InstaUI page")

            ui.server().run_with(app)


            if __name__ == "__main__":
                import uvicorn
                uvicorn.run(app, host="127.0.0.1", port=8080)

        """
        from fastapi import FastAPI
        from instaui.internal.runtime.web.options import WebUserRunWithOptions
        from instaui.internal.runtime.web.web_runtime import WebRuntime

        runtime = WebRuntime(replace(self.launch_config, prefix=prefix))
        assert isinstance(app, FastAPI), "app must be a FastAPI instance"
        opts = WebUserRunWithOptions(
            app=app,
            tags=tags,
            dependencies=dependencies,
            responses=responses,
        )

        runtime.run_with(opts)



from instaui import resources
from instaui.internal.assets.context import GLOBAL_ASSETS

_FRAMEWORK_BOOTSTRAPPED = False


def inject_framework_assets() -> None:
    """Inject framework assets into the assets pool."""

    global _FRAMEWORK_BOOTSTRAPPED
    if _FRAMEWORK_BOOTSTRAPPED:
        return

    GLOBAL_ASSETS.import_maps.update(
        {
            "vue": resources.VUE_ES_JS_PATH,
            "instaui": resources.APP_ES_JS_PATH,
        }
    )

    GLOBAL_ASSETS.css_links.add(resources.APP_CSS_PATH)

    _FRAMEWORK_BOOTSTRAPPED = True



from __future__ import annotations
from typing import Callable
from abc import ABC, abstractmethod
from pathlib import Path


class JsCodegenServices(ABC):
    @abstractmethod
    def path_map_url(self, path: Path) -> str: ...

    @abstractmethod
    def register_watch_handler(self, key: str, fn: Callable, output_count: int): ...

    @abstractmethod
    def register_event_handler(
        self, key: str, fn: Callable, output_count: int, dataset_input_indexs: list[int]
    ): ...

    @abstractmethod
    def resolve_upload_url(self, key: str, handler: Callable) -> str: ...



from .bindable import BindableMixin
from .enums import InputBindingType
from .protocol import CanInputProtocol


class EventContext(CanInputProtocol, BindableMixin):
    def __init__(
        self, path: str, *, input_type: InputBindingType = InputBindingType.EventContext
    ):
        self.path = path
        self._input_type = input_type

    @staticmethod
    def dataset(name="eventData"):
        return EventContext(
            f":e=> e.target.dataset.{name}",
            input_type=InputBindingType.EventContextDataset,
        )

    @staticmethod
    def args():
        return EventContext(":(...e)=> e")

    @staticmethod
    def e():
        return EventContext(":e => e")

    @staticmethod
    def target_value():
        return EventContext(":e => e.target.value")

    def _to_event_input_type(self) -> InputBindingType:
        return self._input_type

    @property
    def _used(self) -> bool:
        return True

    def _mark_used(self) -> None:
        pass

    def _mark_provided(self) -> None:
        pass


class DatasetEventContext(CanInputProtocol):
    def __init__(self, event_context: EventContext) -> None:
        self._event_context = event_context

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.EventContextDataset



from typing import cast
from instaui.internal.ui.path_var import PathInfo


def normalize_paths(path_infos: list[PathInfo]):
    all_binds = all(path.is_bind for path in path_infos)

    if all_binds:
        return [cast(list, path.args)[0] for path in path_infos]

    return [normalize_path_info(path) for path in path_infos]


def normalize_path_info(path_info: PathInfo):
    if path_info.args is None:
        return [path_info.name]

    return [
        path_info.name,
        path_info.args,
    ]



import sys
from instaui.debug.model import SourceSpan
from instaui.debug.source_fallback import _get_source_span_by_module_filter
from .config import is_dev, freeze


# -------------------------
# prod implementation
# -------------------------
class NullSourceSpan:
    __slots__ = ()

    def __repr__(self) -> str:
        return "<NoSourceSpan>"


_NULL_SOURCE_SPAN = NullSourceSpan()


def _get_source_span_prod() -> SourceSpan:
    return _NULL_SOURCE_SPAN  # type: ignore


# -------------------------
# dev implementation
# -------------------------
def _get_source_span_dev():
    import inspect

    frame = inspect.currentframe()
    f = frame.f_back if frame else None

    while f:
        fn = f.f_globals.get(f.f_code.co_name)
        if getattr(fn, "__instaui_user_api__", False):
            f = f.f_back
            assert f is not None

            column = None
            if sys.version_info >= (3, 11):
                column = _get_frame_column(f)

            return SourceSpan(
                file=f.f_code.co_filename,
                line=f.f_lineno,
                function=f.f_code.co_name,
                column=column,
            )
        f = f.f_back

    # fallback
    return _get_source_span_by_module_filter()


def _get_frame_column(frame) -> int | None:
    # Python 3.11+ only
    try:
        code = frame.f_code
        positions = code.co_positions()
        index = frame.f_lasti // 2

        pos = list(positions)[index]
        _, _, col_start, _ = pos
        return col_start
    except Exception:
        return None


# -------------------------
# import-time binding
# -------------------------
if is_dev():
    get_source_span = _get_source_span_dev
else:
    get_source_span = _get_source_span_prod


freeze()



from instaui.internal import import_presets
from instaui.internal.ast.core import AppBootstrap, CustomComponent, Plugin
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.expr_codegen import ExpressionCodegen


class AppBootstrapCodegen:
    def __init__(self, ctx: FileCodegenContext) -> None:
        self.ctx = ctx
        self.import_table = ctx.imports
        self.expr = ExpressionCodegen(ctx)

    def emit(self, bootstrap: AppBootstrap) -> str:
        method = self.import_table.use_from_preset(import_presets.Instaui.install())
        self.ctx.names.declare_top_level_var(bootstrap.target)

        app_var = self.ctx.names.resolve(bootstrap.target)

        parts: list[str] = []
        parts.append(self.parse_app_def(app_var, method, bootstrap))

        if bootstrap.custom_components:
            parts.append(
                self._emit_custom_components(app_var, bootstrap.custom_components)
            )

        if bootstrap.plugins:
            parts.append(self._emit_plugins(app_var, bootstrap.plugins))

        return "\n".join(parts)

    def parse_app_def(self, app_var: str, method: str, bootstrap: AppBootstrap) -> str:
        return f"const {app_var} = {method}({self.ctx.names.resolve(bootstrap.root.id)},{self.expr.emit(bootstrap.meta)})"

    def _emit_plugins(self, app_var: str, plugins: list[Plugin]) -> str:
        parts: list[str] = []
        for plugin in plugins:
            parts.append(self._emit_plugin(app_var, plugin))
        return "\n".join(parts)

    def _emit_plugin(self, app_var: str, plugin: Plugin) -> str:
        args = self.expr.emit(plugin.options) if plugin.options else ""
        return f"{app_var}.use((await import({self.expr.emit(plugin.url)})),{args})"

    def _emit_custom_components(
        self, app_var: str, custom_components: list[CustomComponent]
    ) -> str:
        parts: list[str] = []
        for custom_component in custom_components:
            parts.append(self._emit_custom_component(app_var, custom_component))
        return "\n".join(parts)

    def _emit_custom_component(
        self, app_var: str, custom_component: CustomComponent
    ) -> str:
        return f"{app_var}.component({self.expr.emit(custom_component.name)},(await import({self.expr.emit(custom_component.url)})).default)"



import typing

from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.bindable import mark_used
from instaui.internal.ui.utils.validators import ensure_output_list
from . import watch_types


class JsWatch:
    def __init__(
        self,
        code: str,
        inputs: typing.Optional[typing.Sequence] = None,
        outputs: typing.Optional[typing.Sequence] = None,
        immediate: bool = True,
        deep: typing.Union[bool, int] = False,
        once: bool = False,
        flush: typing.Optional[watch_types.TFlush] = None,
    ) -> None:
        ensure_output_list(outputs)
        self.code = code

        self._inputs = UILiteralExpr.try_parse_list(inputs or [])
        self._outputs = outputs

        self._immediate = immediate
        self._deep = deep
        self._once = once
        self._flush = flush

        mark_used(self._inputs)
        mark_used(self._outputs)


def js_watch(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    outputs: typing.Optional[typing.Sequence] = None,
    code: str,
    immediate: bool = True,
    deep: typing.Union[bool, int] = True,
    once: bool = False,
    flush: typing.Optional[watch_types.TFlush] = None,
):
    """
    Creates a client-side observer that executes JavaScript code in response to reactive source changes.

    Args:
        inputs (typing.Optional[typing.Sequence], optional): Reactive sources to observe. Changes to these sources
                                   trigger the watcher's JavaScript execution.
        outputs (typing.Optional[typing.Sequence], optional): Output targets associated with this watcher. Used for
                                    coordination with other observers.
        code (str, optional): JavaScript code to execute when changes are detected. The code has access
                  to the current values of observed inputs through the `args` parameter.
        immediate (bool, optional):If True, executes the watcher immediately after creation with current values. Defaults to True.
        deep (typing.Union[bool, int], optional): Controls depth of change detection:
                               - True: Recursively tracks nested properties
                               - False: Shallow comparison only
                               - int: Maximum depth level to track (for complex objects).
                               Defaults to True.
        once (bool, optional): If True, automatically stops observation after first trigger. Defaults to False.
        flush (typing.Optional[_types.TFlush], optional): Controls when to flush updates:
                                      - 'sync': Execute immediately on change
                                      - 'post': Batch updates and execute after current tick
                                      - 'pre': Execute before render phase (if applicable)

    # Example:
    .. code-block:: python
        from instaui import ui, html

        num = ui.state(0)
        msg = ui.state('')
        ui.js_watch(inputs=[num], outputs=[msg], code="num => `The number is ${num}`")

        html.number(num)
        ui.text(msg)
    """

    watch = JsWatch(code, inputs, outputs, immediate, deep, once, flush)
    get_current_scope().register_js_watch(watch)
    return watch



from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.bindable import BindableMixin, VarableBindHelper
from instaui.internal.ui.enums import InputBindingType
from instaui.internal.ui.path_var import PathVar
from instaui.internal.ui.protocol import ObservableProtocol
from instaui.internal.ui.variable import Variable


class SlotProp(
    PathVar,
    Variable,
    ObservableProtocol,
    BindableMixin,
):
    def __init__(self) -> None:
        super().__init__()
        define_scope = get_current_scope()

        self._bind_helper = VarableBindHelper(self, define_scope=define_scope)

    def get(self, name: str):
        return self[name]

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _is_providable(self) -> bool:
        return False



from typing import Any, Optional, Union
from instaui.internal.ast import expression
from instaui.internal.pipeline.ast_build.base import ExprBuilderProtocol
from instaui.internal.ui.missing import TMissing
from instaui.internal.ui.pre_setup import PreSetupAction


class PreSetupParser:
    def __init__(self, expr_builder: ExprBuilderProtocol):
        self.expr_builder = expr_builder

    def parse(self, pre_setup: list[list]) -> Optional[expression.ListExpr]:
        if not pre_setup:
            return None

        pre_setup = [
            [
                self.expr_builder.build(target),
                self.parse_item_value(running_value),
                self.parse_item_value(reset_value),
            ]
            for target, running_value, reset_value in pre_setup
        ]

        return expression.ListExpr(
            [
                self.expr_builder.list_shallow(pre_setup_item)
                for pre_setup_item in pre_setup
            ]
        )

    def parse_item_value(
        self, maybe_action: Union[Any, TMissing]
    ) -> expression.Expression:
        if isinstance(maybe_action, TMissing):
            return expression.ObjectLiteral(
                {
                    "type": "missing",
                }
            )

        if isinstance(maybe_action, PreSetupAction):
            return self.expr_builder.object_shallow(
                {
                    "type": "action",
                    "value": expression.ObjectExpr.from_kwargs(
                        **{
                            "inputs": self.expr_builder.list_shallow(
                                maybe_action.inputs
                            )
                            or None,
                            "code": expression.RawFunctionExpr(maybe_action.code),
                        }
                    ),
                }
            )

        return self.expr_builder.object_shallow(
            {"type": "const", "value": self.expr_builder.build(maybe_action)}
        )



from typing import Any
from instaui.systems.dataclass_system import dataclass, field


@dataclass()
class JsExpr:
    code: str


@dataclass(frozen=True)
class ReprKey:
    value: Any = field(hash=True)



from enum import Enum
from pathlib import Path
from typing import Callable, Optional, Union
from instaui.systems.dataclass_system import dataclass


class PageBuildStrategy(Enum):
    LAZY = "lazy"
    EAGER = "eager"


@dataclass(frozen=True)
class WebViewUserLaunchConfig:
    assets_path: Path
    debug: bool
    title: str = "InstaUI"
    auto_create_window: Union[bool, str] = "/"
    on_page_mounted: Optional[Callable] = None
    clean_assets_on_start: bool = True
    page_build_strategy: PageBuildStrategy = PageBuildStrategy.LAZY
    prebuild_routes: Optional[list[str]] = None



from instaui.internal.ast.core import App
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.expr_codegen import ExpressionCodegen
from .component_codegen import ComponentCodegen


class AppCodegen:
    def __init__(self, ctx: FileCodegenContext) -> None:
        self.ctx = ctx
        self.names = ctx.names
        self.import_table = ctx.imports
        self.bindings = ctx.root.binding_registry
        self.app_bootstrap_codegen = ctx.root.components.app_bootstrap_codegen_factory(
            ctx
        )
        self.expr_codegen = ExpressionCodegen(ctx)

    def emit(self, app: App) -> str:
        component_codegen = ComponentCodegen(self.ctx)

        parts: list[str] = []

        # component name registry
        for comp in app.components:
            self.names.declare_component(comp.id)

        # components
        for comp in app.components:
            parts.append(component_codegen.emit(comp))

        parts.append(self.app_bootstrap_codegen.emit(app.bootstrap))

        body = "\n\n".join(parts)
        return f"{self.import_table.render()}\n\n{body}"



from __future__ import annotations
from typing import cast

from instaui.constants.ui import INSTAUI_JS_MODULE_NAME, USE_DARK_REF_METHOD
from instaui.debug.api_boundary import user_api
from instaui.internal.ui.custom_var import CustomVar


@user_api
def use_dark() -> bool:
    """
    On start up, it reads the value from localStorage/sessionStorage (the key is configurable) to see if there is a user configured color scheme, if not, it will use users' system preferences.


    Example:
    .. code-block:: python
        from instaui import ui,html

        @ui.page('/')
        def index():
            dark = ui.use_dark()
            html.checkbox(dark)
    """

    return cast(
        bool,
        CustomVar(INSTAUI_JS_MODULE_NAME, USE_DARK_REF_METHOD),
    )



from __future__ import annotations
from typing import Callable
from instaui.systems.dataclass_system import dataclass, field


@dataclass()
class CodegenResult:
    code: str
    watch_handlers: list[WatchHandler] = field(default_factory=list)
    event_handlers: list[EventHandler] = field(default_factory=list)
    file_upload_handlers: list[FileUploadHandler] = field(default_factory=list)


@dataclass()
class WatchHandler:
    key: str
    fn: Callable
    output_count: int


@dataclass()
class EventHandler:
    key: str
    fn: Callable
    output_count: int
    dataset_input_indexs: list[int] = field(default_factory=list)


@dataclass()
class FileUploadHandler:
    key: str
    fn: Callable



from html import escape
from typing import Optional


def render_attrs(attrs: Optional[dict]) -> str:
    if not attrs:
        return ""

    parts: list[str] = []

    for k, v in attrs.items():
        if v is False or v is None:
            continue
        elif v is True:
            parts.append(k)
        else:
            parts.append(f'{k}="{escape(str(v), quote=True)}"')

    return " " + " ".join(parts)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Protocol
from instaui.protocol.module_import.import_item import PresetProtocol

if TYPE_CHECKING:
    from instaui.internal.ast.core import AppBootstrap


class ImportTableProtocol(Protocol):
    def use(
        self,
        *,
        module: str,
        name: str,
        alias: Optional[str] = None,
    ) -> str: ...

    def use_from_preset(
        self,
        preset: PresetProtocol,
    ) -> str: ...

    def render(self) -> str: ...


class AppBootstrapCodegenProtocol(Protocol):
    def emit(self, bootstrap: AppBootstrap) -> str: ...



from __future__ import annotations
from typing import Union
from instaui.internal.ui.element import Element
from instaui.internal.ui.value_element import ValueElement
from ._mixins import InputEventMixin


_T_value = Union[int, float, None]


class Number(InputEventMixin, ValueElement[_T_value]):
    def __init__(
        self,
        value: _T_value = None,
        *,
        model_value: _T_value = None,
        min: _T_value = None,
        max: _T_value = None,
    ):
        super().__init__("input", value, is_html_component=True)
        self.props({"type": "number"})

        if min is not None:
            self.props({"min": min})
        if max is not None:
            self.props({"max": max})
        if model_value is not None:
            self.props({"value": model_value})

    def _input_event_mixin_element(self) -> Element:
        return self



from typing import Callable
from instaui.internal.ui.bindable import BindableMixin


class UploadEndpoint(BindableMixin):
    def __init__(self, handler: Callable) -> None:
        self.handler = handler

    @property
    def _used(self) -> bool:
        return True

    def _mark_used(self) -> None:
        pass

    def _mark_provided(self) -> None:
        pass



from instaui.constants.runtime import RuntimeMode
from instaui.internal.context.page_context import in_page_context
from instaui.internal.ui.app_context import get_app


class RuntimeCtx:
    def _app(self):
        assert in_page_context(), "Should be used inside ui.page() context"
        app = get_app()
        return app

    @property
    def mode(self):
        return self._app().mode

    @property
    def is_web(self) -> bool:
        return self.mode == RuntimeMode.WEB

    @property
    def is_webview(self) -> bool:
        return self.mode == RuntimeMode.WEBVIEW

    @property
    def is_zero(self) -> bool:
        return self.mode == RuntimeMode.ZERO

    @property
    def is_debug(self):
        return self._app().debug_mode


runtime_ctx = RuntimeCtx()



from __future__ import annotations
from typing import TYPE_CHECKING, Protocol


if TYPE_CHECKING:
    from instaui.internal.ui.renderable import Renderable
    from instaui.internal.ui._scope import Scope
    from instaui.internal.ast.core import Render, ComponentRef, ComponentDef
    from instaui.internal.ast.expression import Expression, ListExpr, ObjectExpr
    from instaui.internal.pipeline.ast_build.expr_builder import TBuildMode


class RenderBuilderProtocol(Protocol):
    def build(self, renderable: Renderable) -> Render: ...


class ExprBuilderProtocol(Protocol):
    def build(self, value: Expression, mode: TBuildMode = "expr") -> Expression: ...

    def list_shallow(self, values: list) -> ListExpr: ...

    def object_shallow(self, kwargs: dict) -> ObjectExpr: ...


class ComponentBuilderProtocol(Protocol):
    def build_component_ref(self, scope: Scope) -> ComponentRef: ...

    def build_component_def(self, scope: Scope) -> ComponentDef: ...



from instaui.internal.ast.core import App
from instaui.internal.codegen.context import CodegenRootContext, FileCodegenContext
from .file_unit import FileUnit
from .component_codegen import ComponentCodegen
from .component_file_map import ComponentFileMap


class FileCodegen:
    def __init__(
        self, file_map: ComponentFileMap, root_ctx: CodegenRootContext
    ) -> None:
        self.ctx = FileCodegenContext(root_ctx)
        self._file_map = file_map
        self.app_bootstrap_codegen = root_ctx.components.app_bootstrap_codegen_factory(
            self.ctx
        )

    # ================================
    # public API
    # ================================
    def emit(self, unit: FileUnit, app: App) -> str:
        """
        Generate complete code for a single file
        """

        # --------------------------------
        # Cross-file component import
        # --------------------------------
        # self._emit_component_imports(unit)

        # --------------------------------
        # component codegen
        # --------------------------------
        body_chunks: list[str] = []
        component_codegen = ComponentCodegen(self.ctx)

        for comp in unit.components:
            body_chunks.append(component_codegen.emit(comp))

        # --------------------------------
        # entry file
        # --------------------------------
        if unit.is_entry:
            body_chunks.append(self.app_bootstrap_codegen.emit(app.bootstrap))

        # --------------------------------
        # Assemble final code
        # --------------------------------

        import_code = self.ctx.imports.render()
        return f"{import_code}\n\n{''.join(body_chunks)}"

    # ================================
    # helpers
    # ================================

    def _emit_component_imports(self, unit: FileUnit) -> None:
        """
        Scan ComponentRef and generate imports for cross-file components
        """
        pass
        # collector = ComponentRefCollector()
        # refs: set[ComponentRef] = set()

        # for comp in unit.components:
        #     for render in comp.renders:
        #         # collector.collect(render, refs)
        #         pass

        # for ref in refs:
        #     target_path = self._file_map.resolve(ref.id)

        #     # Same file component, no import needed
        #     if target_path == unit.path:
        #         continue

    def _relative_path(self, from_path: str, to_path: str) -> str:
        """
        Calculate relative path (minimum viable implementation)
        """
        # components/Foo.js -> components/Bar.js
        from_dir = from_path.rsplit("/", 1)[0]
        to_file = to_path.rsplit("/", 1)[-1]
        name = to_file.removesuffix(".js")

        if from_dir == to_path.rsplit("/", 1)[0]:
            return f"./{name}"

        return f"./{to_path}"



import inspect
from typing import Any, Callable, Literal, Optional, Sequence
from instaui.internal.ast import expression
from instaui.internal.ui.enums import InputBindingType
from instaui.internal.ui.input_slient_data import InputSilentData
from instaui.internal.ui.protocol import (
    CanInputProtocol,
    CanOutputProtocol,
    ObservableProtocol,
)

_SYNC_TYPE = "sync"
_ASYNC_TYPE = "async"


def normalize_f_type(f_type: Optional[Literal["async", "sync"]]):
    if f_type is None or f_type == "sync":
        return None

    return f_type


def inputs_to_types(inputs: Sequence[Any]):
    return [
        input._to_event_input_type().value
        if isinstance(input, CanInputProtocol)
        else InputBindingType.Data.value
        for input in inputs
    ]


def outputs_to_types(outputs: Sequence[CanOutputProtocol]):
    return [output._to_event_output_type().value for output in outputs]


def analyze_observable_inputs(
    inputs: Sequence,
) -> tuple[list, list[int], list[int]]:
    """
    Returns:
        inputs, slients, datas
    """

    slients: list[int] = [0] * len(inputs)
    datas: list[int] = [0] * len(inputs)
    result_inputs = []

    for idx, input in enumerate(inputs):
        if isinstance(input, ObservableProtocol):
            result_inputs.append(input)
            continue

        elif isinstance(input, CanInputProtocol):
            slients[idx] = 1
            result_inputs.append(
                input.value if isinstance(input, InputSilentData) else input
            )

        elif input is None:
            result_inputs.append(expression.UNDEFINED)
            datas[idx] = 1

        else:
            datas[idx] = 1
            result_inputs.append(input)

    return result_inputs, slients, datas


def py_function_type(fn: Callable):
    return _ASYNC_TYPE if inspect.iscoroutinefunction(fn) else _SYNC_TYPE


def auto_made_inputs_to_slient(
    inputs: list,
    outputs: list,
):
    if (not inputs) or (not outputs):
        return inputs

    outputs_set = set(outputs)

    return [
        InputSilentData(input) if input in outputs_set else input for input in inputs
    ]


def normalize_int_values(values: Optional[Sequence[int]], by: int = 0):
    if not values:
        return None

    if all(x == by for x in values):
        return None

    return values



from __future__ import annotations
from typing import Any, Mapping, Sequence
from instaui.systems.dataclass_system import dataclass


@dataclass(eq=False)
class UILiteralExpr:
    value: Any

    @classmethod
    def try_parse(cls, value: Any):
        if (
            isinstance(value, (int, float, str, bool, list, dict, tuple))
            or value is None
        ):
            return UILiteralExpr(value)

        return value

    @staticmethod
    def try_parse_list(value: Sequence) -> list:
        return [UILiteralExpr.try_parse(x) for x in value]

    @staticmethod
    def try_parse_dict(value: Mapping) -> dict:
        return {k: UILiteralExpr.try_parse(v) for k, v in value.items()}



from typing import Any, cast
import webview
from instaui.internal.backend.base import BaseBackend
from instaui.internal.backend.common import python_handler_utils
from instaui.internal.backend.invocation import BackendInvocation, BackendInvocationKind
from instaui.internal.runtime.webview.config import WebViewUserLaunchConfig
from instaui.internal.runtime.webview.options import WebViewUserRunOptions
from instaui.protocol.invocation.spec import ComputedSpec, EventSpec
from .watch_handler import register_watch_handler
from .event_handler import register_event_handler
from . import api


class WebViewBackend(BaseBackend):
    def __init__(self, config: WebViewUserLaunchConfig) -> None:
        self.launch_config = config

    def create_window(
        self,
        url: str,
    ):
        """Create a new window. Returns the window object of pywebview."""
        window = webview.create_window(self.launch_config.title, url, js_api=api.Api())

        if self.launch_config.on_page_mounted:

            def on_app_mounted():
                self.launch_config.on_page_mounted(window)  # type: ignore

            window.expose(on_app_mounted)

        return window

    def run(self, run_options: WebViewUserRunOptions):
        webview.start(**run_options.webview_start_args, debug=self.launch_config.debug)

    def _run_with_index(
        self, index_url: str, run_options: WebViewUserRunOptions | None
    ):
        webview.create_window(self.launch_config.title, index_url, js_api=api.Api())

        opts = {}
        if run_options:
            opts.update(run_options.webview_start_args)

        opts.update({"debug": self.launch_config.debug})

        webview.start(**opts)

    def register_invocation(self, ref_id: int, invocation: BackendInvocation) -> Any:
        route = invocation.render_ctx.route
        assert route is not None, "route must be set in render_ctx for invocation"
        key = python_handler_utils.create_handler_key(
            route, invocation.fn, self.launch_config.debug
        )

        if (
            invocation.kind == BackendInvocationKind.COMPUTED
            or invocation.kind == BackendInvocationKind.WATCH
        ):
            spec = cast(ComputedSpec, invocation.spec)
            register_watch_handler(
                key,
                invocation.fn,
                spec.outputs_binding_count,
                spec.custom_type_adapter_map,
            )

            return key

        if invocation.kind == BackendInvocationKind.EVENT:
            spec = cast(EventSpec, invocation.spec)
            register_event_handler(
                key,
                invocation.fn,
                spec.outputs_binding_count,
                spec.dataset_input_indexs,
                spec.custom_type_adapter_map,
            )

            return key

        if invocation.kind == BackendInvocationKind.FILE_UPLOAD:
            raise NotImplementedError
            # file_upload_router.register_upload_file_handler(
            #     key, invocation.fn, self.endpoints.UPLOAD_URL_WITH_PREFIX
            # )

            # return f"{self.endpoints.UPLOAD_URL_WITH_PREFIX}?hkey={key}"



import typing
from typing_extensions import ParamSpec
from instaui.debug.api_boundary import user_api
from instaui.debug.model import SourceSpan
from instaui.debug.source import get_source_span
from instaui.internal.codec import build_type_adapter_map
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_index import new_scope_if_needed
from instaui.internal.ui.event_context import DatasetEventContext
from instaui.internal.ui.pre_setup import normalize_pre_setup
from instaui.protocol.invocation.spec import EventSpec

from .variable import Variable
from .event_modifier import TEventModifier
from .app_context import get_app, get_current_scope
from .event import EventMixin
from .utils.validators import ensure_output_list
from .bindable import BindableMixin, VarableBindHelper, mark_used

P = ParamSpec("P")
R = typing.TypeVar("R")


class WebEvent(EventMixin, BindableMixin, Variable, typing.Generic[P, R]):
    def __init__(
        self,
        fn: typing.Callable[P, R],
        inputs: typing.Sequence,
        outputs: typing.Sequence,
        pre_setup: typing.Optional[list] = None,
        modifier: typing.Optional[typing.Sequence[TEventModifier]] = None,
        *,
        _source_span_: SourceSpan,
    ):
        self._source_span_ = _source_span_
        new_scope_if_needed()
        ensure_output_list(outputs)
        define_scope = get_current_scope()
        self._inputs = UILiteralExpr.try_parse_list(inputs)
        self._outputs = outputs
        self._fn = fn
        self._pre_setup = normalize_pre_setup(pre_setup or [])
        self._modifier = modifier or []

        self._bind_helper = VarableBindHelper(
            self,
            define_scope.register_web_event,
            define_scope=define_scope,
            lazy_mark_used=[
                *self._inputs,
                *self._outputs,
                *[x[0] for x in self._pre_setup],
            ],
        )

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self._fn(*args, **kwargs)

    def _attach_to_element(
        self,
        *,
        extends: typing.Optional[typing.Sequence],
        modifier: typing.Optional[typing.Sequence[TEventModifier]] = None,
    ):
        mark_used(extends)
        mark_used(self)

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _export_invocation_spec(self) -> EventSpec:
        type_adapter_map = build_type_adapter_map(self._fn, self._inputs)
        dataset_input_indexs = [
            i
            for i, input_ in enumerate(self._inputs or [])
            if isinstance(input_, DatasetEventContext)
        ]
        return EventSpec(
            len(self._outputs or []), dataset_input_indexs, type_adapter_map
        )


@user_api
def event(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    outputs: typing.Optional[typing.Sequence] = None,
    pre_setup: typing.Optional[list] = None,
):
    """
    Creates an event handler decorator for binding reactive logic to component events.

    Args:
        inputs (typing.Optional[typing.Sequence], optional): Reactive sources (state objects, computed properties)
                                   that should be accessible during event handling.
                                   These values will be passed to the decorated function
                                   when the event fires.
        outputs (typing.Optional[typing.Sequence], optional): Targets (state variables, UI elements) that should
                                    update when this handler executes. Used for coordinating
                                    interface updates after the event is processed.
        pre_setup (typing.Optional[list], optional): A list of pre-setup actions to be executed before the event executes.


    # Example:
    .. code-block:: python
        from instaui import ui, html

        a = ui.state(0)

        @ui.event(inputs=[a], outputs=[a])
        def plus_one(a):
            return a + 1

        html.button("click me").on_click(plus_one)
        html.paragraph(a)

    use pre_setup:
    .. code-block:: python
        a = ui.state(0)
        task_running = ui.state(False)

        @ui.event(inputs=[a], outputs=[a], pre_setup=[task_running,True,False])
        async def long_running_task(a):
            await asyncio.sleep(3)
            return a + 1

        html.button("click me").on_click(long_running_task).disabled(task_running)
    """

    if get_app().mode == "zero":
        raise Exception(
            "Cannot use event decorator in zero mode. You should use `ui.js_event` instead."
        )

    _source_span_ = get_source_span()

    def wrapper(func: typing.Callable[P, R]):
        return WebEvent(
            func,
            inputs or [],
            outputs=outputs or [],
            pre_setup=pre_setup,
            _source_span_=_source_span_,
        )

    return wrapper



from instaui.internal.ui.app_context import get_current_container
from instaui.internal.ui.bindable import mark_used
from instaui.internal.ui.renderable import Renderable
from instaui.internal.ui.container import Container


class VIf(Renderable, Container):
    def __init__(self, on: bool | int | str):
        """
        A conditional container that renders its children only when the given condition is True.

        Args:
            on (bool | int | str): A boolean or reactive reference that determines whether the container's
                                child elements should be displayed. If True, children are rendered;
                                if False, children are hidden.


        Example:
        .. code-block:: python
            from instaui import ui, html

            value = ui.state(False)

            html.button("toggle").on_click(
                ui.js_event(inputs=[value], outputs=[value], code="(v)=> !v")
            )

            with ui.vif(value):
                ui.text("show")
        """

        super().__init__()
        get_current_container().add_child(self)

        mark_used(on)
        self._on = on
        self._renderables: list[Renderable] = []

    def add_child(self, renderable: Renderable):
        self._renderables.append(renderable)



__all__ = [
    "element",
    "convert_reference",
    "register_plugin",
    "scope_context",
    "CustomVar",
    "StateReadableCustomVar",
    "CssAsset",
    "CssRole",
    "on_page_exit",
    "register_component_extension",
    "refs",
    "page_once",
    "runtime",
    "RuntimeMode",
]

from instaui.internal.ui.element import Element as element
from instaui.internal.ui.reference import convert_reference
from instaui.internal.assets import register_plugin
from instaui.internal.ui._scope import Scope as scope_context
from instaui.internal.ui.custom_var import CustomVar, StateReadableCustomVar
from instaui.internal.assets.css_assets import CssAsset, CssRole
from instaui.internal.context.page_context import on_page_exit
from instaui.internal.assets import register_component_extension
from instaui.internal.ui.props_injectable import refs
from .page_once import page_once
from ._runtime import runtime_ctx as runtime
from instaui.constants.runtime import RuntimeMode



from typing import Generic, TypeVar
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.bindable import BindableMixin, VarableBindHelper
from instaui.internal.ui.enums import InputBindingType, OutputSetType
from instaui.internal.ui.path_var import PathVar
from instaui.internal.ui.protocol import (
    CanOutputProtocol,
    ObservableProtocol,
)
from instaui.internal.ui.variable import Variable


_T = TypeVar("_T")


class VForItem(
    Variable,
    PathVar,
    CanOutputProtocol,
    ObservableProtocol,
    BindableMixin,
    Generic[_T],
):
    def __init__(self):
        super().__init__()
        self._bind_helper = VarableBindHelper(self, define_scope=get_current_scope())

    def __getattr__(self, name: str):
        return self[name]

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.Ref

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _is_providable(self) -> bool:
        return False


class VForIndex(
    Variable,
    PathVar,
    ObservableProtocol,
    BindableMixin,
):
    def __init__(self):
        super().__init__()
        self._bind_helper = VarableBindHelper(self, define_scope=get_current_scope())

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _is_providable(self) -> bool:
        return False


class VForItemKey(Variable, PathVar, ObservableProtocol, BindableMixin):
    def __init__(self):
        super().__init__()
        self._bind_helper = VarableBindHelper(self, define_scope=get_current_scope())

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _is_providable(self) -> bool:
        return False


TVForItem = VForItem
TVForIndex = VForIndex



from __future__ import annotations
from typing import TYPE_CHECKING
from instaui.internal.ast import expression
from instaui.internal.pipeline.ast_build.component_builder import ComponentBuilder
from instaui.internal.pipeline.ast_build.custom_component_builder import (
    CustomComponentBuilder,
)
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder
from instaui.internal.pipeline.ast_build.plugins_builder import PluginsBuilder
from instaui.internal.pipeline.ast_build.variable_decl_builder import (
    VariableDeclBuilder,
)

from instaui.internal.ast import core as ast_core
from instaui.internal.pipeline.ast_build.context import (
    AstBuilderContext,
    AstBuildInput,
)

if TYPE_CHECKING:
    from instaui.internal.runtime.runtime_ctx import RuntimeContext
    from instaui.internal.runtime.render_ctx import RenderContext
    from instaui.internal.pipeline.asset_resolver.resolver import AssetResolver


class AstBuilder:
    def __init__(
        self,
        runtime_ctx: RuntimeContext,
        render_ctx: RenderContext,
        asset_resolver: AssetResolver,
    ) -> None:
        self.runtime_ctx = runtime_ctx
        self.ctx = AstBuilderContext(
            runtime_ctx.binding_registry, render_ctx, asset_resolver
        )
        self._variable_builder = VariableDeclBuilder(self.ctx)
        self._expr_builder = ExprBuilder(self.ctx)
        self._component_builder = ComponentBuilder(runtime_ctx, render_ctx, self.ctx)
        self._plugins_builder = PluginsBuilder(runtime_ctx, render_ctx, self.ctx)
        self._custom_component_builder = CustomComponentBuilder(
            runtime_ctx, render_ctx, self.ctx
        )

    def build(self, ast_input: AstBuildInput) -> ast_core.App:
        app = ast_input.app
        scopes = ast_input.scopes

        assert scopes, "App scope not set"

        root = scopes[0]
        root_ref = self._component_builder.build_component_ref(root)

        components = []

        for scope in app.created_scopes:
            components.append(self._component_builder.build_component_def(scope))

        return ast_core.App(
            root=root_ref,
            components=components,
            bootstrap=ast_core.AppBootstrap(
                self.ctx.get_app_id(),
                root_ref,
                meta=expression.ObjectLiteral(self.ctx.render_ctx.app_meta),
                plugins=self._plugins_builder.build(list(ast_input.assets.plugins))
                if ast_input.assets.plugins
                else None,
                custom_components=self._custom_component_builder.build(
                    ast_input.assets.component_dependencies
                )
                if ast_input.assets.component_dependencies
                else None,
            ),
        )



from __future__ import annotations
import threading
from typing import TYPE_CHECKING, Callable, Optional
from fastapi import APIRouter, Request, UploadFile, Query, File
from starlette.datastructures import UploadFile as StarletteUploadFile
from .systems import async_system
from instaui.internal.backend.common import handler as _handler


if TYPE_CHECKING:
    from .endpoints import ServerEndpoints


_event_handlers: dict[str, _handler.HandlerInfo] = {}
dict_lock = threading.Lock()


def create_router(router: APIRouter, endpoints: ServerEndpoints):
    @router.post(endpoints.UPLOAD_URL)
    async def _(
        request: Request,
        hkey: str = Query(...),
        files: list[UploadFile] = File(None),
        file: UploadFile = File(None),
    ):
        handler = _get_handler(hkey)
        if handler is None:
            raise ValueError("event handler not found")

        real_file = await _convert_file(request, files, file)
        return await async_system.maybe_async(handler.fn, real_file)


def _get_handler(hkey: str):
    return _event_handlers.get(hkey)


async def _convert_file(
    request: Request,
    files: Optional[list[UploadFile]] = None,
    file: Optional[UploadFile] = None,
):
    if file:
        return file
    if files:
        return files

    form = await request.form()
    indexed_files = []
    for key, value in form.multi_items():
        if key.startswith("file["):
            if isinstance(value, StarletteUploadFile):
                indexed_files.append(value)

    return indexed_files


def register_upload_file_handler(key: str, handler: Callable, upload_url: str):
    upload_url = f"{upload_url}?hkey={key}"

    if key in _event_handlers:
        return upload_url

    handler_info = _handler.HandlerInfo.from_handler(
        handler,
        0,
        skip_convert_param=True,
    )

    with dict_lock:
        _event_handlers[key] = handler_info

    return upload_url



from __future__ import annotations
from typing import Any
from instaui.internal.ui.bindable import (
    BindableMixin,
    mark_used,
    is_bindable,
    mark_provided,
)


class UIMaybeRefExpr(BindableMixin):
    def __init__(self, value: Any) -> None:
        self.value = value

    @property
    def _used(self) -> bool:
        return self.value._used if is_bindable(self.value) else True

    def _mark_used(self) -> None:
        mark_used(self.value)

    def _mark_provided(self) -> None:
        if is_bindable(self.value):
            mark_provided(self.value)



from instaui.internal import import_presets
from instaui.internal.ast.core import ComponentDef, ExprWatchCall, JsWatchCall
from instaui.internal.ast import expression
from instaui.internal.ast.symbol import SymbolId
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.expr_codegen import ExpressionCodegen
from instaui.internal.codegen.render_codegen import RenderCodegen
from instaui.internal.codegen.variable_decl_codegen import VariableDeclCodegen
from instaui.internal.codegen.web_watch_task import WebWatchTaskRenderer


class ComponentCodegen:
    def __init__(self, ctx: FileCodegenContext) -> None:
        self.ctx = ctx
        self.import_table = ctx.imports
        self.bindings = ctx.root.binding_registry
        self.expr = ExpressionCodegen(ctx)
        self.var_decl = VariableDeclCodegen(ctx)
        self._web_watch_task = WebWatchTaskRenderer(ctx)

    def emit(self, comp: ComponentDef) -> str:
        # declare variables
        for var in comp.vars:
            self.ctx.names.declare_var(var)

        name = self.ctx.names.resolve(comp.id)

        parts: list[str] = []

        # setup
        setup_lines: list[str] = []

        # injects
        if comp.injects:
            setup_lines.append(self.emit_injects(comp.injects))

        # variables
        setup_lines.extend(self.var_decl.emit(var) for var in comp.variables)

        # web watch tasks
        if comp.web_watch_tasks:
            setup_lines.append(self._web_watch_task.emit(comp.web_watch_tasks))

        # js watch
        if comp.js_watchs:
            setup_lines.append(self.emit_js_watch_call(comp.js_watchs))

        # expr watch
        if comp.expr_watchs:
            setup_lines.append(self.emit_expr_watch_call(comp.expr_watchs))

        if comp.on_mounted_calls:
            mounted_method = self.import_table.use_from_preset(
                import_presets.Vue.on_mounted()
            )
            setup_lines.extend(
                [
                    f"{mounted_method}(() => {{",
                    "\n    ".join(
                        self.expr.emit(call) for call in comp.on_mounted_calls
                    ),
                    "});",
                ]
            )

        # provide
        if comp.provides:
            setup_lines.append(self.emit_provides(comp.provides))

        # render
        render_codegen = RenderCodegen(self.ctx)
        render_code = render_codegen.emit_renders(comp.renders)

        setup_body = "\n    ".join(setup_lines)
        parts.append(
            f"""
const {name} = {{
  setup() {{
    {setup_body}
    return () => {render_code};
  }}
}};
""".strip()
        )

        return "\n\n".join(parts)

    def emit_injects(self, injects: list[SymbolId]) -> str:
        method = self.import_table.use_from_preset(import_presets.Instaui.inject())
        names = [self.ctx.names.resolve(inject) for inject in injects]
        return rf"const {{{', '.join(names)}}} = {method}();"

    def emit_provides(self, provides: list[SymbolId]) -> str:
        method = self.import_table.use_from_preset(import_presets.Instaui.provide())
        names = [self.ctx.names.resolve(provide) for provide in provides]
        return rf"{method}({{{', '.join(names)}}});"

    def emit_js_watch_call(self, watchs: list[JsWatchCall]) -> str:
        method = self.import_table.use_from_preset(import_presets.Instaui.js_watch())
        args_list = expression.ListExpr([watch.args for watch in watchs])
        return rf"{method}({self.expr.emit(args_list)});"

    def emit_expr_watch_call(self, watchs: list[ExprWatchCall]) -> str:
        method = self.import_table.use_from_preset(import_presets.Instaui.expr_watch())
        args_list = expression.ListExpr([watch.args for watch in watchs])
        return rf"{method}({self.expr.emit(args_list)});"



from __future__ import annotations
from collections.abc import Hashable
from typing import Any, TypeVar, Callable, Generic, Optional, Sequence, Union
from typing_extensions import ParamSpec
from instaui.debug.api_boundary import user_api
from instaui.debug.model import SourceSpan
from instaui.debug.source import get_source_span
from instaui.internal.codec import build_type_adapter_map
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.pre_setup import normalize_pre_setup
from instaui.internal.ui.utils.validators import ensure_output_list
from instaui.protocol.invocation.spec import WatchSpec

from .bindable import mark_used
from .app_context import get_app, get_current_scope
from . import watch_types


P = ParamSpec("P")
R = TypeVar("R")


class WebWatch(Generic[P, R]):
    def __init__(
        self,
        func: Callable[P, R],
        inputs: Optional[Sequence] = None,
        outputs: Optional[Sequence] = None,
        immediate: bool = True,
        deep: Union[bool, int] = True,
        once: bool = False,
        flush: Optional[watch_types.TFlush] = None,
        pre_setup: Optional[list] = None,
        _debug: Optional[Any] = None,
        extra_key: Optional[Sequence[Hashable]] = None,
        *,
        _source_span_: SourceSpan,
    ) -> None:
        self._source_span_ = _source_span_
        ensure_output_list(outputs)
        self._inputs = UILiteralExpr.try_parse_list(inputs or [])
        self._outputs = outputs

        self._fn = func
        self._immediate = immediate
        self._deep = deep
        self._once = once
        self._flush = flush
        self._debug = _debug
        self._pre_setup = normalize_pre_setup(pre_setup or [])
        self._extra_key = extra_key

        mark_used(self._inputs)
        mark_used(self._outputs)
        mark_used([x[0] for x in self._pre_setup or []])

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self._fn(*args, **kwargs)

    def _export_invocation_spec(self) -> WatchSpec:
        type_adapter_map = build_type_adapter_map(self._fn, self._inputs)
        return WatchSpec(len(self._outputs or []), type_adapter_map)


@user_api
def watch(
    *,
    inputs: Optional[Sequence] = None,
    outputs: Optional[Sequence] = None,
    immediate: bool = True,
    deep: Union[bool, int] = True,
    once: bool = False,
    flush: Optional[watch_types.TFlush] = None,
    pre_setup: Optional[list] = None,
    extra_key: Optional[Sequence[Hashable]] = None,
    _debug: Optional[Any] = None,
):
    """
    Creates an observer that tracks changes in reactive sources and triggers callbacks.

    Args:
        inputs (Optional[Sequence], optional): Reactive sources to observe (state objects or computed properties).
                                   Changes to these sources trigger the watcher callback.
        outputs (Optional[Sequence], optional): Output targets associated with this watcher.
                                    Used for coordination with computed properties or other observers.
        immediate (bool, optional): If True, executes callback immediately after creation with current values. Defaults to True.
        deep (Union[bool, int], optional): Controls depth of change detection:
                               - True: Recursively tracks nested properties
                               - False: Shallow comparison only
                               - int: Maximum depth level to track (for complex objects).
                               Defaults to True.
        once (bool, optional):  If True, automatically stops observation after first trigger. Defaults to False.
        flush (Optional[watch_types.TFlush], optional): Controls when to flush updates:
                                      - 'sync': Execute immediately on change
                                      - 'post': Batch updates and execute after current tick
                                      - 'pre': Execute before render phase (if applicable)
        pre_setup (Optional[list], optional): A list of pre-setup actions to be executed before the event executes.


    # Example:
    .. code-block:: python
        from instaui import ui, html

        num = ui.state(0)
        msg = ui.state('')

        @ui.watch(inputs=[num], outputs=[msg])
        def when_num_change(num):
            return f"The number is {num}"

        html.number(num)
        ui.text(msg)

    list append:
    .. code-block:: python
        from instaui import ui, html

        num = ui.state(0)
        msg = ui.state([])

        @ui.watch(inputs=[num, msg], outputs=[msg])
        def when_num_change(num, msg:list):
            msg.append(f"The number changed to {num}")
            return msg

        html.number(num)
        ui.text(msg)

    """
    app = get_app()
    if app.mode == "zero":
        raise Exception(
            "Cannot use watch decorator in zero mode. You should use `ui.js_watch` instead."
        )

    _source_span_ = get_source_span()

    def wrapper(func: Callable[P, R]):
        obj = WebWatch(
            func,
            inputs,
            outputs=outputs,
            immediate=immediate,
            deep=deep,
            once=once,
            flush=flush,
            pre_setup=pre_setup,
            _debug=_debug,
            extra_key=extra_key,
            _source_span_=_source_span_,
        )

        get_current_scope().register_web_watch(obj)

    return wrapper



import inspect
from instaui.debug.model import SourceSpan

_INTERNAL_MODULE_PREFIXES = ("instaui.",)


_INTERNAL_PATH_KEYWORDS = (
    "/instaui/",
    "\\instaui\\",
)


def _is_internal_frame(frame) -> bool:
    module = frame.f_globals.get("__name__", "")
    if any(module.startswith(p) for p in _INTERNAL_MODULE_PREFIXES):
        return True

    filename = frame.f_code.co_filename
    if any(k in filename for k in _INTERNAL_PATH_KEYWORDS):
        return True

    return False


def _get_source_span_by_module_filter() -> SourceSpan:
    frame = inspect.currentframe()
    if frame is None:
        return _unknown_source()

    f = frame.f_back

    while f is not None:
        if not _is_internal_frame(f):
            return SourceSpan(
                file=f.f_code.co_filename,
                line=f.f_lineno,
                function=f.f_code.co_name,
            )
        f = f.f_back

    return _unknown_source()


def _unknown_source() -> SourceSpan:
    return SourceSpan(
        file="<unknown>",
        line=0,
        function="<unknown>",
    )



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Optional, Union
from typing_extensions import Self
from instaui.internal.ui.protocol import ObservableProtocol, CanOutputProtocol
from instaui.systems.dataclass_system import dataclass, field
from .bindable import VarableBindHelper, BindableMixin
from .app_context import get_current_scope
from .enums import InputBindingType, OutputSetType


@dataclass
class OperatorStep:
    op: str
    value: Any


@dataclass
class UnaryStep:
    op: str


TPathStep = str | int | OperatorStep | UnaryStep

BIND_FLAG = "b"


@dataclass()
class PathInfo:
    name: str
    args: Optional[list[Any]] = field(default=None)
    is_bind: Optional[bool] = field(init=False)

    def __post_init__(self):
        self.is_bind = self.name == BIND_FLAG


class PathableMixin(ABC):
    @abstractmethod
    def not_(self) -> bool:
        pass

    @abstractmethod
    def len_(self) -> int:
        pass


class PathVar(PathableMixin):
    def __getattr__(self, key):
        return self[key]

    def __getitem__(self, item: Union[str, int]):
        return PathTrackerBindable(self)[item]

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    def not_(self):
        return PathTrackerBindable(self).not_()

    def __add__(self, other: str):
        return PathTrackerBindable(self) + other

    def __radd__(self, other: str):
        return other + PathTrackerBindable(self)

    def __sub__(self, other: Union[int, float]):
        return PathTrackerBindable(self) - other

    def __rsub__(self, other: Union[int, float]):
        return other - PathTrackerBindable(self)

    def __mul__(self, other: Union[int, float]):
        return PathTrackerBindable(self) * other

    def __rmul__(self, other: Union[int, float]):
        return other * PathTrackerBindable(self)

    def __truediv__(self, other: Union[int, float]):
        return PathTrackerBindable(self) / other

    def __rtruediv__(self, other: Union[int, float]):
        return other / PathTrackerBindable(self)

    def __and__(self, other: Any):
        return other & PathTrackerBindable(self)

    def __or__(self, other: Any):
        return other | PathTrackerBindable(self)

    def __lt__(self, other):
        return PathTrackerBindable(self) < other

    def __le__(self, other):
        return PathTrackerBindable(self) <= other

    def __gt__(self, other):
        return PathTrackerBindable(self) > other

    def __ge__(self, other):
        return PathTrackerBindable(self) >= other

    def __ne__(self, other):
        return PathTrackerBindable(self) != other

    def len_(self):
        return PathTrackerBindable(self).len_()


class PathTracker(PathableMixin):
    def __init__(self, paths: Optional[list[TPathStep]] = None):
        self._paths = paths or []

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    def __getitem__(self, key) -> Self:
        return self.__new_self__([*self._paths, key])

    def __getattr__(self, key) -> Self:
        return self.__new_self__([*self._paths, key])

    def not_(self) -> Self:
        return self.__new_self__([*self._paths, UnaryStep("!")])

    def __add__(self, other: str) -> Self:
        return self.__new_self__([*self._paths, OperatorStep("+", other)])

    def __radd__(self, other: str) -> Self:
        return self.__new_self__([*self._paths, OperatorStep("~+", other)])

    def __sub__(self, other: Union[int, float]):
        return self.__new_self__([*self._paths, OperatorStep("-", other)])

    def __rsub__(self, other: Union[int, float]):
        return self.__new_self__([*self._paths, OperatorStep("~-", other)])

    def __mul__(self, other: Union[int, float]):
        return self.__new_self__([*self._paths, OperatorStep("*", other)])

    def __rmul__(self, other: Union[int, float]):
        return self.__new_self__([*self._paths, OperatorStep("~*", other)])

    def __truediv__(self, other: Union[int, float]):
        return self.__new_self__([*self._paths, OperatorStep("/", other)])

    def __rtruediv__(self, other: Union[int, float]):
        return self.__new_self__([*self._paths, OperatorStep("~/", other)])

    def __and__(self, other: Any):
        return self.__new_self__([*self._paths, OperatorStep("&&", other)])

    def __or__(self, other: Any):
        return self.__new_self__([*self._paths, OperatorStep("||", other)])

    def __lt__(self, other):
        return self.__new_self__([*self._paths, OperatorStep("<", other)])

    def __le__(self, other):
        return self.__new_self__([*self._paths, OperatorStep("<=", other)])

    def __gt__(self, other):
        return self.__new_self__([*self._paths, OperatorStep(">", other)])

    def __ge__(self, other):
        return self.__new_self__([*self._paths, OperatorStep(">=", other)])

    def __ne__(self, other):
        return self.__new_self__([*self._paths, OperatorStep("!=", other)])

    def len_(self):
        return self.__new_self__([*self._paths, "length"])

    @abstractmethod
    def __new_self__(self, paths: list[TPathStep]) -> Self:
        pass


class PathTrackerBindable(
    PathTracker, BindableMixin, CanOutputProtocol, ObservableProtocol
):
    def __init__(self, source: PathableMixin):
        super().__init__()
        self._source = source
        define_scope = get_current_scope()

        self._bind_helper = VarableBindHelper(
            self,
            define_scope=define_scope,
            lazy_mark_used=[
                source,
                lambda: [
                    path.value if isinstance(path, OperatorStep) else path
                    for path in self._paths
                    if not isinstance(path, UnaryStep)
                ],
            ],
            maybe_provides=[source],
        )

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    def __new_self__(self, paths: list[Union[str, list[str]]]) -> PathTrackerBindable:
        obj = PathTrackerBindable(self._source)
        obj._paths = paths
        return obj

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.Ref

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()


class PathTrackerPipelines:
    @staticmethod
    def to_expr_literal(path_tracker: PathTrackerBindable) -> list:
        return [PathStepPipelines.to_expr_literal(path) for path in path_tracker._paths]


class PathStepPipelines:
    @staticmethod
    def to_expr_literal(step: TPathStep):
        if isinstance(step, OperatorStep):
            return {"op": step.op, "v": step.value}

        if isinstance(step, UnaryStep):
            return {"op": step.op}

        return step



from typing import Optional, cast
from instaui.constants.ui import INSTAUI_JS_MODULE_NAME, USE_PAGE_TITLE_METHOD
from instaui.debug.api_boundary import user_api
from instaui.internal.ui.custom_var import CustomVar


@user_api
def use_page_title(title: Optional[str] = None) -> str:
    """Set the title of the HTML document.

    Args:
        title (str): The title of the HTML document.
    """

    if isinstance(title, CustomVar):
        return title

    return cast(
        str, CustomVar(INSTAUI_JS_MODULE_NAME, USE_PAGE_TITLE_METHOD, args=(title,))
    )



from typing import cast
from typing_extensions import Self
from .element import Element


class DisabledElement:
    def disabled(self, disabled: bool = True) -> Self:
        return cast(Element, self).props({"disabled": disabled})  # type: ignore



from __future__ import annotations
import inspect
import threading
from typing import TYPE_CHECKING, Callable, Mapping, Optional
from fastapi import Response, APIRouter
from instaui.internal.backend.common import handler as _handler
from instaui.internal.backend.common.event_dataset import event_dataset_type_adapter
from instaui.protocol.codec.type_adapter import TypeAdapterProtocol
from . import _utils

if TYPE_CHECKING:
    from .endpoints import ServerEndpoints


_event_handlers: dict[str, _handler.HandlerInfo] = {}
dict_lock = threading.Lock()


def create_router(router: APIRouter, debug_mode: bool, endpoints: ServerEndpoints):
    _async_handler(router, endpoints)
    _sync_handler(router, endpoints)

    if debug_mode:

        @router.get("/instaui/event-infos", tags=["instaui-debug"])
        def event_infos():
            return _get_statistics_info()


def _async_handler(router: APIRouter, endpoints: ServerEndpoints):
    @router.post(endpoints.ASYNC_EVENT_URL)
    async def _(data: dict, response: Response):
        handler = _get_handler(data)
        if handler is None:
            raise _utils.HandlerNotFoundError("event handler not found")

        assert inspect.iscoroutinefunction(handler.fn), (
            "handler must be a coroutine function"
        )

        result = await handler.fn(*handler.get_handler_args(_get_binds_from_data(data)))
        return _utils.response_web_data(handler.outputs_binding_count, result, response)


def _sync_handler(router: APIRouter, endpoints: ServerEndpoints):
    @router.post(endpoints.EVENT_URL)
    def _(data: dict, response: Response):
        handler = _get_handler(data)
        if handler is None:
            raise _utils.HandlerNotFoundError("event handler not found")

        result = handler.fn(*handler.get_handler_args(_get_binds_from_data(data)))

        return _utils.response_web_data(handler.outputs_binding_count, result, response)


def register_event_handler(
    key: str,
    handler: Callable,
    outputs_binding_count: int,
    dataset_input_indexs: list[int],
    custom_type_adapter_map: Optional[Mapping[int, TypeAdapterProtocol]] = None,
):
    if key in _event_handlers:
        return

    custom_type_adapter_map = dict(custom_type_adapter_map or {}) | {
        i: event_dataset_type_adapter for i in dataset_input_indexs
    }

    with dict_lock:
        _event_handlers[key] = _handler.HandlerInfo.from_handler(
            handler,
            outputs_binding_count,
            custom_type_adapter_map=custom_type_adapter_map,
        )


def _get_handler(data: dict):
    return _event_handlers.get(data["hKey"])


def _get_binds_from_data(data: dict):
    return [bind for bind in data.get("bind", [])]


def _get_statistics_info():
    return {
        "_event_handlers count": len(_event_handlers),
        "_event_handlers keys": list(_event_handlers.keys()),
    }



from __future__ import annotations
from typing import Optional, Literal, Union

from instaui.internal.ui.element import Element
from ._responsive_type._common import (
    TMaybeResponsive,
    TLevel_1_9,
)
from ._responsive_type._typography import (
    TWeightEnum,
    TTextWrapEnum,
    TTrimEnum,
    TAlignEnum,
)


class Heading(Element):
    def __init__(
        self,
        text: Optional[str] = None,
        *,
        as_: Optional[Literal["h1", "h2", "h3", "h4", "h5", "h6"]] = "h1",
        as_child: Optional[bool] = None,
        size: Optional[TMaybeResponsive[TLevel_1_9]] = "6",
        weight: Optional[TMaybeResponsive[Union[TWeightEnum, str]]] = None,
        align: Optional[TMaybeResponsive[Union[TAlignEnum, str]]] = None,
        trim: Optional[Union[TTrimEnum, str]] = None,
        truncate: Optional[bool] = None,
        text_wrap: Optional[Union[TTextWrapEnum, str]] = None,
    ):
        """
        Creates a heading element with customizable HTML tag level and typography styles.

        Args:
            text (Optional[TMaybeRef[str]]): The text content of the heading. Can be a static string or a reactive reference.
            as_ (Optional[TMaybeRef[Literal["h1", "h2", "h3", "h4", "h5", "h6"]]]): Specifies the HTML heading tag level.
                Defaults to "h1".
            as_child (Optional[TMaybeRef[bool]]): If True, renders the heading component without wrapping its own element,
                allowing it to inherit the parent context.
            size (Optional[TMaybeResponsive[TLevel_1_9]]): Controls the font size using a predefined size scale.
                Defaults to "6".
            weight (Optional[TMaybeResponsive[Union[TWeightEnum, str]]]): Sets the font weight style.
            align (Optional[TMaybeResponsive[Union[TAlignEnum, str]]]): Sets text alignment such as left,
                center, or right.
            trim (Optional[TMaybeRef[Union[TTrimEnum, str]]]): Adjusts whitespace trimming behavior.
            truncate (Optional[TMaybeRef[bool]]): If True, truncates long text with ellipsis when overflow occurs.
            text_wrap (Optional[TMaybeRef[Union[TTextWrapEnum, str]]]): Controls text wrapping behavior.

        Example:
        .. code-block:: python
            from instaui import ui

            # Render a default h1 heading
            ui.heading("ui.heading")
        """

        super().__init__("heading")

        self.props(
            {
                "innerText": text,
                "as": as_,
                "as_child": as_child,
                "size": size,
                "weight": weight,
                "text_align": align,
                "trim": trim,
                "truncate": truncate,
                "text_wrap": text_wrap,
            }
        )



from contextlib import contextmanager
from typing import Iterator, Optional
from instaui.internal.ast.core import ComponentDef
from instaui.internal.ast.symbol import SymbolId


class ComponentScope:
    def __init__(self) -> None:
        self._stack: list[ComponentDef] = []

    @contextmanager
    def push(self, comp: ComponentDef) -> Iterator[None]:
        self._stack.append(comp)
        try:
            yield
        finally:
            self._stack.pop()

    @property
    def current(self) -> Optional[ComponentDef]:
        if not self._stack:
            return None
        return self._stack[-1]

    def declare_var(self, symbol: SymbolId) -> None:
        comp = self.current
        if comp is None:
            raise RuntimeError("Variable declared outside component")

        comp.vars.append(symbol)



from __future__ import annotations
from typing import Optional
from instaui.internal.ast.base import AstNode
from instaui.internal.ast.expression import (
    Expression,
    IdentifierRef,
    JsonLiteralExpr,
    ListLiteral,
    ObjectExpr,
    ObjectLiteral,
    RawFunctionExpr,
    Literal,
    StringLiteral,
)
from instaui.internal.ast.symbol import SymbolId
from instaui.internal.ast.sfc_template import TemplateElementInfo
from instaui.systems.dataclass_system import dataclass, field


@dataclass()
class App:
    root: ComponentRef
    components: list[ComponentDef]
    bootstrap: AppBootstrap


@dataclass()
class SvgSprite:
    args: ObjectLiteral


@dataclass()
class AppBootstrap:
    target: SymbolId
    root: ComponentRef
    meta: Expression
    plugins: Optional[list[Plugin]] = None
    custom_components: Optional[list[CustomComponent]] = None


@dataclass()
class Plugin:
    url: StringLiteral
    options: Optional[ObjectExpr | JsonLiteralExpr]


@dataclass()
class CustomComponent:
    name: StringLiteral
    url: StringLiteral


@dataclass()
class ComponentRef:
    id: SymbolId
    vfor: Optional[ComponentRefVFor] = None


@dataclass()
class ComponentRefVFor:
    index: list[SymbolId] = field(default_factory=list)
    value: list[SymbolId] = field(default_factory=list)
    item_key: list[SymbolId] = field(default_factory=list)

    def __bool__(self) -> bool:
        return bool(self.index or self.value or self.item_key)


@dataclass()
class Render:
    pass


@dataclass()
class ComponentRefRender(Render):
    ref: ComponentRef


@dataclass()
class Element(Render):
    tag: Expression
    props: Optional[ElementProp] = None
    classes: Optional[ElementClasses] = None
    styles: Optional[ElementStyles] = None
    events: Optional[ObjectExpr] = None
    dirs: Optional[list[Directive]] = None
    lifecycles: Optional[dict[str, Expression]] = None
    slots: Optional[dict[str, Slot]] = None

    # normalize_for_template
    _tpl: Optional[TemplateElementInfo] = None


@dataclass()
class VFor(Render):
    fkey: Literal
    array: VForArray
    used_item: Optional[SymbolId] = None
    used_index: Optional[SymbolId] = None
    used_key: Optional[SymbolId] = None
    transition_group: Optional[ObjectLiteral] = None
    children: Optional[list[Render]] = None


@dataclass()
class MatchRender(Render):
    condition: Expression
    cases: list[MatchCase]
    default_case: Optional[list[Render]] = None


@dataclass()
class MatchCase:
    value: Expression
    children: list[Render]


@dataclass()
class ContentRender:
    content: Expression


@dataclass()
class VIf(Render):
    condition: Expression
    children: Optional[list[Render]] = None


@dataclass()
class VForArray:
    type: Literal
    value: Expression


@dataclass()
class Slot:
    """
    One slot = one render fragment.
    It will ultimately be codegen'ed into () => [...].
    """

    name: StringLiteral  # "default" / "header" / ...
    body: list[Render]
    used_prop: Optional[SymbolId] = None


@dataclass()
class ElementProp:
    binding: Optional[dict[str, Expression]] = None
    proxy: Optional[list[Expression]] = None
    static: Optional[JsonLiteralExpr] = None
    ref: Optional[IdentifierRef] = None

    def __bool__(self):
        return bool(self.binding or self.proxy or self.static or self.ref)


@dataclass()
class ElementClasses:
    binding: Optional[list[Expression]] = None
    static: Optional[ListLiteral] = None
    maps: Optional[ObjectExpr] = None

    def __bool__(self):
        return bool(self.binding or self.static or self.maps)


@dataclass()
class ElementStyles:
    static: Optional[JsonLiteralExpr] = None
    binding: Optional[ObjectExpr] = None
    proxy: Optional[list[Expression]] = None

    def __bool__(self):
        return bool(self.binding or self.static or self.proxy)


@dataclass()
class Directive:
    name: Literal
    value: Expression
    sys: Optional[Literal] = None
    arg: Optional[Literal] = None
    mf: Optional[ListLiteral] = None


@dataclass()
class ComponentDef:
    id: SymbolId

    variables: list[VariableDecl] = field(default_factory=list)
    renders: list[Render] = field(default_factory=list)
    web_watch_tasks: Optional[list[WebWatchTask]] = None
    js_watchs: Optional[list[JsWatchCall]] = None
    expr_watchs: Optional[list[ExprWatchCall]] = None
    on_mounted_calls: list[RawFunctionExpr] = field(default_factory=list)
    injects: Optional[list[SymbolId]] = None
    provides: Optional[list[SymbolId]] = None

    vars: list[SymbolId] = field(default_factory=list)


@dataclass()
class WebWatchTask(AstNode):
    fn: BackendFunctionRef
    args: ObjectExpr


@dataclass()
class JsWatchCall:
    args: ObjectExpr


@dataclass()
class ExprWatchCall:
    args: ObjectExpr


@dataclass()
class VariableDecl(AstNode):
    target: SymbolId


@dataclass()
class ConstDataDecl(VariableDecl):
    value: Expression


@dataclass()
class ElementRefDecl(VariableDecl):
    pass


@dataclass()
class ExprComputedDecl(VariableDecl):
    args: ObjectExpr


@dataclass()
class ExprEventDecl(VariableDecl):
    args: ObjectExpr


@dataclass()
class RefDecl(VariableDecl):
    args: ObjectExpr


@dataclass()
class JsComputedDecl(VariableDecl):
    args: ObjectExpr


@dataclass()
class JsEventDecl(VariableDecl):
    args: ObjectExpr


@dataclass()
class JsFnDecl(VariableDecl):
    fn: RawFunctionExpr


@dataclass()
class CustomRefDecl(VariableDecl):
    module: str
    name: str
    params: Optional[list[Expression]] = None


@dataclass()
class BackendComputedDecl(VariableDecl):
    args: ObjectExpr


@dataclass()
class BackendEventDecl(VariableDecl):
    args: ObjectExpr


@dataclass()
class BackendFunctionRef:
    ref_id: int



from __future__ import annotations
from typing import Union
from instaui.internal.ui.element import Element
from instaui.internal.ui.components.content import Content
from instaui.internal.ui.vfor import VFor


class Table(Element):
    def __init__(
        self,
        columns: Union[list[str], None] = None,
        rows: Union[list, None] = None,
    ):
        """Create a table element.

        Args:
            columns (Union[list[str], None], optional): A list of column headers or a reactive reference to such a list. Defaults to None.
            rows (Union[list, None], optional): A list of row data, where each row is a list of cell values, or a reactive reference to such a list. Defaults to None.
        """
        super().__init__("table")

        with self:
            with Element("thead"), Element("tr"):
                with VFor(columns) as col:  # type: ignore
                    with Element("th"):
                        Content(col)

            with Element("tbody"):
                with VFor(rows) as row:  # type: ignore
                    with Element("tr"):
                        with VFor(row) as cell:  # type: ignore
                            with Element("td"):
                                Content(cell)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
from instaui.internal.assets.snapshot import ComponentDependencySnapshot
from instaui.internal.pipeline.ast_build.component_builder import ComponentBuilder
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder
from instaui.internal.ast import core as ast_core
from instaui.internal.ast import expression
from instaui.internal.pipeline.ast_build.context import AstBuilderContext

if TYPE_CHECKING:
    from instaui.internal.runtime.runtime_ctx import RuntimeContext
    from instaui.internal.runtime.render_ctx import RenderContext


class CustomComponentBuilder(ComponentBuilder):
    def __init__(
        self,
        runtime_ctx: RuntimeContext,
        render_ctx: RenderContext,
        builder_ctx: AstBuilderContext,
    ) -> None:
        self.runtime_ctx = runtime_ctx
        self.ctx = builder_ctx
        self._expr_builder = ExprBuilder(self.ctx)
        self._component_builder = ComponentBuilder(runtime_ctx, render_ctx, self.ctx)

    def build(
        self, components: Sequence[ComponentDependencySnapshot]
    ) -> Optional[list[ast_core.CustomComponent]]:
        if not components:
            return None

        return [
            ast_core.CustomComponent(
                name=expression.StringLiteral(comp.tag),
                url=expression.StringLiteral(
                    self.ctx.asset_resolver.resolve_asset(comp.esm)
                ),
            )
            for comp in components
        ]



from __future__ import annotations
from typing import TYPE_CHECKING
from instaui.systems.dataclass_system import dataclass

if TYPE_CHECKING:
    from instaui.internal.ast import expression


class PropertyKey:
    pass


@dataclass(frozen=True)
class StringKey(PropertyKey):
    value: str


@dataclass(frozen=True)
class ExpressionKey(PropertyKey):
    expr: expression.Expression



from __future__ import annotations
from enum import Enum, auto
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Generic, Optional, TypeVar
from instaui.protocol.invocation.spec import InvocationSpec

if TYPE_CHECKING:
    from instaui.internal.runtime.render_ctx import RenderContext


SpecT = TypeVar("SpecT", bound=InvocationSpec)


class BackendInvocationKind(Enum):
    COMPUTED = auto()
    WATCH = auto()
    EVENT = auto()
    FILE_UPLOAD = auto()


@dataclass(frozen=True)
class BackendInvocation(Generic[SpecT]):
    kind: BackendInvocationKind
    fn: Callable
    render_ctx: RenderContext
    spec: Optional[SpecT] = None
    source: Optional[str] = None  # debug info



from __future__ import annotations
from typing import TYPE_CHECKING, Union

from instaui.internal.ui.value_element import ValueElement
from ._mixins import InputEventMixin

if TYPE_CHECKING:
    from instaui.internal.ui.element import Element

_T_value = str


class Date(InputEventMixin, ValueElement[_T_value]):
    def __init__(
        self,
        value: Union[_T_value, None] = None,
        *,
        model_value: Union[_T_value, None] = None,
    ):
        super().__init__("input", value, is_html_component=True)
        self.props({"type": "date"})

        if model_value is not None:
            self.props({"value": model_value})

    def _input_event_mixin_element(self) -> Element:
        return self  # type: ignore



from __future__ import annotations
from typing import Optional, Union
from instaui.internal.ui.element import Element
from instaui.internal.ui.value_element import ValueElement
from instaui.internal.ui.disabled_element import DisabledElement
from ._mixins import InputEventMixin


class Textarea(InputEventMixin, DisabledElement, ValueElement[str]):
    def __init__(
        self,
        value: Union[str, None] = None,
        *,
        model_value: Union[str, None] = None,
        disabled: Optional[bool] = None,
    ):
        super().__init__("textarea", value, is_html_component=True)

        if disabled is not None:
            self.props({"disabled": disabled})
        if model_value is not None:
            self.props({"value": model_value})

    def _input_event_mixin_element(self) -> Element:
        return self



from typing import Callable, Generator, Optional, Union, overload
from instaui.internal.context.page_context import in_page_context
from instaui.internal.ui.app_context import get_default_app


TLayoutFn = Callable[[], Generator[None, None, None]]


class Layout:
    def __init__(self, fn: TLayoutFn, priority: int = 0) -> None:
        self.fn = fn
        self.priority = priority

    def create_generator(self) -> Generator[None, None, None]:
        return self.fn()


@overload
def layout(fn: TLayoutFn) -> TLayoutFn: ...


@overload
def layout(*, priority: int) -> Callable[[TLayoutFn], TLayoutFn]: ...


def layout(
    fn: Optional[TLayoutFn] = None, *, priority: int = 0
) -> Union[TLayoutFn, Callable[[TLayoutFn], TLayoutFn]]:
    """
    Define a global page layout.

    The decorated function MUST be a generator function
    and MUST yield exactly once.
    """

    def decorator(fn: TLayoutFn) -> TLayoutFn:
        assert not in_page_context(), (
            "layout() should be called outside ui.page() context"
        )
        app = get_default_app()
        app.append_layout(Layout(fn, priority=priority))
        return fn

    if fn is None:
        return decorator
    else:
        return decorator(fn)


def clear_layout():
    assert not in_page_context(), (
        "remove_layout() should be called outside ui.page() context"
    )
    app = get_default_app()
    app.layouts.clear()



from typing import Any, Optional
from instaui.internal.ast.symbol import SymbolId


class IdGenerator:
    def __init__(self, start_id: int = 1) -> None:
        self._next_uid = start_id

    def next_id(self) -> int:
        uid = self._next_uid
        self._next_uid += 1
        return uid


class SymbolMapper:
    def __init__(self, id_generator: Optional[IdGenerator] = None) -> None:
        self._id_generator = id_generator or IdGenerator()
        self._mapping: dict[Any, SymbolId] = {}

    def get_or_create_id(self, obj: Any) -> SymbolId:
        sid = self._mapping.get(obj)
        if sid is None:
            sid = self._new_symbol()
            self._mapping[obj] = sid
        return sid

    def get_id(self, obj: Any) -> Optional[SymbolId]:
        return self._mapping.get(obj)

    def _new_symbol(self) -> SymbolId:
        uid = self._id_generator.next_id()
        return SymbolId(uid)



from typing import Any, Callable, TypeVar, cast

from instaui.systems.dataclass_system import dataclass, field
from instaui.internal.ui.app_context import get_app
from instaui.internal.ui.upload import UploadEndpoint
from ._protocols import UploadFileProtocol

THandlerArgs = TypeVar("THandlerArgs", UploadFileProtocol, list[UploadFileProtocol])


@dataclass(frozen=True)
class UploadFileEvent:
    _endpoint: UploadEndpoint = field(init=True)

    @property
    def url(self) -> str:
        return cast(str, self._endpoint)


def upload_file():
    """
    Decorator that registers a file upload handler and returns an upload event object.

    Args:
        handler (Callable[[UploadFileProtocol], Any]): A function that processes uploaded files.
            Takes an UploadFileProtocol object as input and returns any serializable result.

    Returns:
        UploadFileEvent: An event object containing the URL endpoint for file uploads.

    Example:
    .. code-block:: python
        import shutil
        from pathlib import Path
        from instaui import ui, file_io

        @file_io.upload_file()
        def upload_file(file: file_io.TUploadFile) -> file_io.TUploadFileResult:
            save_path = Path(__file__).parent / file.filename

            # read content in chunks(suitable for large files)
            with save_path.open("wb") as f:
                shutil.copyfileobj(file.file, f)

            # read all content(suitable for small files)
            # content = await file.read()
            # save_path.write_bytes(content)

            return {"status": 200, "files": [str(save_path.absolute())]}

        # The upload_file.url can be used in JavaScript to handle file uploads
        print(upload_file.url)
    """

    def wrapper(
        handler: Callable[[THandlerArgs], Any],
    ):
        app = get_app()
        if app.mode == "zero":
            raise Exception("Cannot use upload_file decorator in zero mode.")

        endpoint = UploadEndpoint(handler=handler)
        return UploadFileEvent(endpoint)

    return wrapper



from pathlib import Path
from typing import Callable, Optional, Union
from instaui.cdn.options import CdnResourceOption


class ZeroOptions:
    def __init__(
        self,
        debug: bool = False,
        icons_svg_path: Optional[Union[Path, Callable[[], Path]]] = None,
        cdn_resource_options: Optional[list[CdnResourceOption]] = None,
    ) -> None:
        self.debug = debug
        self.icons_svg_path = icons_svg_path
        self.cdn_resource_options = cdn_resource_options or []

    def get_import_maps_cdn_overrides(self) -> dict[str, str]:
        if not self.cdn_resource_options:
            return {}

        return dict(
            item
            for option in self.cdn_resource_options
            for item in (option.import_maps or {}).items()
        )

    def get_css_links_cdn_overrides(self) -> dict[Path, str]:
        if not self.cdn_resource_options:
            return {}

        return dict(
            item
            for option in self.cdn_resource_options
            for item in (option.css_links or {}).items()
        )



import asyncio
from typing import Any, Awaitable, Callable, TypeVar, Union
from fastapi.concurrency import run_in_threadpool

T = TypeVar("T")


async def maybe_async(
    func: Callable[..., Union[T, Awaitable[T]]], *args: Any, **kwargs: Any
) -> T:
    """
    Run a callable that may be synchronous or asynchronous.

    - If `func` is an async function (defined with `async def`), it will be awaited directly.
    - If `func` is a regular (sync) function, it will be executed in a thread pool
      using `fastapi.concurrency.run_in_threadpool()` to avoid blocking the event loop.

    Args:
        func: The function to call. Can be synchronous or asynchronous.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The result of calling `func`, regardless of whether it's sync or async.

    Example:
        >>> async def async_add(x, y): return x + y
        >>> def sync_add(x, y): return x + y
        >>> await maybe_async(async_add, 1, 2)
        3
        >>> await maybe_async(sync_add, 1, 2)
        3
    """
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    return await run_in_threadpool(func, *args, **kwargs)  # type: ignore[return-value]



from instaui.internal.ui.element import Element


def teleport(
    to: str,
    *,
    defer: bool = True,
    disabled: bool = False,
):
    """
    Teleports UI elements to a target container, allowing dynamic repositioning
    within the page layout.

    Args:
        to (str): The CSS selector or reactive reference specifying
                            the target container where the content should be moved.
        defer (bool, optional): Whether to defer the teleportation until the
                                next render cycle. Defaults to True.
        disabled (bool, optional): A flag or reactive reference to
                                            conditionally disable teleportation.
                                            Defaults to False.

    Example:
    .. code-block:: python
        from instaui import ui, html

        # Basic teleport to a target element
        with ui.column().classes("target-box"):
            ui.content("Container")

        with ui.teleport(".target-box"):
            ui.text("Teleported content")

        # Dynamic teleport with reactive target
        target = ui.state(".box1")
        with ui.teleport(to=target):
            ui.text("Dynamic content")

        # Disabled teleport
        disabled = ui.state(False)
        with ui.teleport(".target", disabled=disabled):
            ui.text("Conditionally teleported")
    """

    ele = Element(tag="teleport")
    ele.props({"to": to})

    if defer is not True:
        ele.props({"defer": False})

    if disabled is not False:
        ele.props({"disabled": disabled})

    return ele



from __future__ import annotations
from pathlib import Path
from typing import Callable, Optional, Union
from instaui import __version__ as _INSTA_VERSION
from instaui import resources
from instaui.constants.runtime import RuntimeMode
from instaui.internal.assets.context import GLOBAL_ASSETS, get_active_assets
from instaui.internal.assets.snapshot import AssetsSnapshot
from instaui.internal.ast.expression import RawFunctionExpr
from instaui.internal.codegen.program_codegen import ProgramCodegen
from instaui.internal.pipeline.ast_build.builder import AstBuilder
from instaui.internal.pipeline.ast_build.context import AstBuildInput
from instaui.internal.pipeline.zero_html.html_renderer import (
    HtmlOptions,
    ZeroHtmlRenderer,
)
from instaui.internal.runtime.app_meta import AppMeta
from instaui.internal.runtime.assets_init import inject_framework_assets
from instaui.internal.runtime.layout import LayoutContext
from instaui.internal.runtime.render_ctx import RenderContext
from instaui.internal.runtime.runtime_ctx import RuntimeContext
from instaui.internal.ui._app import App
from instaui.internal.ui.app_context import get_default_app
from instaui.internal.ui.app_index import new_app
from instaui.internal.pipeline.asset_resolver.zero_resolver import ZeroAssetResolver
from .options import ZeroOptions
from .assets_pipeline import ZERO_ASSETS_PIPELINE


class ZeroRuntime:
    def __init__(self, options: Optional[ZeroOptions] = None):
        self._options = options or ZeroOptions()
        self.asset_resolver = ZeroAssetResolver()
        self.mode = RuntimeMode.ZERO
        self._app_meta = AppMeta(self.mode, _INSTA_VERSION, self._options.debug)

        self.ctx = RuntimeContext()

    def run(self, page_fn: Callable, *, output_path: Optional[Path] = None):
        layout_ctx = LayoutContext(get_default_app().layouts)

        with new_app(self.mode, debug=self._options.debug) as app:
            with layout_ctx:
                page_fn()

            page_assets = get_active_assets()
            inject_framework_assets()
            assets_snapshot = ZERO_ASSETS_PIPELINE.run(GLOBAL_ASSETS, page_assets)

        ast = AstBuilder(
            self.ctx, RenderContext(self._app_meta.to_dict()), self.asset_resolver
        ).build(self._create_ast_build_input(app, assets_snapshot))
        ast.components[0].on_mounted_calls.append(
            RawFunctionExpr(
                r"document.querySelector('zero-loadding').setAttribute('hidden', '')"
            )
        )

        code = ProgramCodegen(self.ctx).emit(ast)
        renderer = ZeroHtmlRenderer()
        html_str = renderer.render(
            code, self._build_render_options(assets_snapshot, self._options)
        )
        if output_path:
            output_path.write_text(html_str, encoding="utf-8")

        return html_str

    def _build_render_options(
        self, merged_assets: AssetsSnapshot, options: ZeroOptions
    ) -> HtmlOptions:
        import_maps = (
            merged_assets.import_maps | self._options.get_import_maps_cdn_overrides()
        )

        return HtmlOptions(
            favicon_url=self._normalize_asset_path(
                merged_assets.favicon or resources.FAVICON_PATH
            ),
            css_links=[
                self._normalize_asset_path(css) for css in merged_assets.css_links
            ],
            js_links=[self._normalize_asset_path(js) for js in merged_assets.js_links],
            import_map={
                name: self._normalize_asset_path(link)
                for name, link in import_maps.items()
            },
            style_tags=[
                (style_tag.group_id, style_tag.content)
                for style_tag in merged_assets.style_tags
            ],
            script_tags=[
                (script_tag.script_attrs, script_tag.content)
                for script_tag in merged_assets.script_tags
            ],
            icons_svg_path=options.icons_svg_path,
        )

    def _normalize_asset_path(self, path: Union[str, Path, None]) -> str:
        if path is None:
            return ""
        if isinstance(path, str):
            return path

        return self.asset_resolver.resolve_asset(path)

    def _create_ast_build_input(
        self, app: App, assets_pool: AssetsSnapshot
    ) -> AstBuildInput:
        return AstBuildInput(app=app, scopes=app.created_scopes, assets=assets_pool)



from instaui.debug.source import get_source_span
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.app_index import new_scope_if_needed
from instaui.internal.ui.bindable import BindableMixin, VarableBindHelper
from instaui.internal.ui.enums import InputBindingType, OutputSetType
from instaui.internal.ui.protocol import CanOutputProtocol, ObservableProtocol
from instaui.internal.ui.variable import Variable


class ElementRef(
    Variable,
    ObservableProtocol,
    CanOutputProtocol,
    BindableMixin,
):
    def __init__(self) -> None:
        """
        A reference holder for a frontend DOM element, enabling two-way interaction between
        Python and JavaScript. The referenced element can be used as input/output in JS
        events, and methods can be executed dynamically on the element at runtime.

        Args:
            None: The reference is created without initialization parameters. It will be
                bound to an element using ``element_ref(...)`` during UI construction.

        Example:
        .. code-block:: python
            # 1️ Use ElementRef inside ui.event as an output
            cp = ui.element_ref()

            @ui.event(outputs=[cp])
            def on_click():
                return ui.run_element_method("reset")

            Counter("my counter").element_ref(cp)
            html.button("reset").on_click(on_click)


            # 2️⃣ Use ElementRef as input to ui.js_event
            ele = ui.element_ref()
            click_event = ui.js_event(inputs=[ele], code="(ele)=> ele.increment()")

            CounterWithRefInput().element_ref(ele)
            html.button("click").on_click(click_event)


            # 3️⃣ Access DOM element inside JavaScript event using `ele.$el ?? ele`
            ele_dom = ui.element_ref()
            output = ui.state("")

            ui.text("custom").element_ref(ele_dom).on(
                "click",
                ui.js_event(
                    inputs=[ele_dom],
                    outputs=[output],
                    code="ele => (ele.$el ?? ele).tagName"
                ),
            )
            ui.text(output)
        """
        self._source_span_ = get_source_span()
        new_scope_if_needed()
        define_scope = get_current_scope()
        self._bind_helper = VarableBindHelper(
            self, define_scope.register_element_ref, define_scope=define_scope
        )

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.ElementRefAction

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.ElementRef


def run_element_method(method_name: str, *args, **kwargs):
    return {
        "method": method_name,
        "args": args,
    }



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING, Union
from instaui.systems.dataclass_system import dataclass


if TYPE_CHECKING:
    from instaui.internal.runtime.zero_runtime import ZeroRenderResult


@dataclass()
class ReportModel:
    result: ZeroRenderResult

    @property
    def html_size_mb(self) -> float:
        return len(self.result.html.encode("utf-8")) / 1024 / 1024

    @property
    def import_maps_records(self):
        import_maps = (
            self.result.assets.import_maps
            | self.result.options.get_import_maps_cdn_overrides()
        )

        return [
            (
                name,
                str(url) if isinstance(url, Path) else url,
                self._try_get_file_size_mb(url) or "-",
            )
            for name, url in import_maps.items()
        ]

    @property
    def css_links(self):
        return [
            (
                str(link),
                self._path_exists_class(link),
                self._try_get_file_size_mb(link) or "-",
            )
            for link in self.result.assets.css_links
        ]

    @property
    def js_links(self):
        return [
            (
                str(link),
                self._path_exists_class(link),
                self._try_get_file_size_mb(link) or "-",
            )
            for link in self.result.assets.js_links
            if isinstance(link, Path) and link.is_file()
        ]

    @property
    def vue_app_component(self):
        result = []

        for info in self.result.assets.component_dependencies:
            is_file = isinstance(info.esm, Path) and info.esm.is_file()
            class_name = str(self._path_exists_class(info.esm) if is_file else "")
            url = str(info.esm) if is_file else "not file"
            size = self._try_get_file_size_mb(info.esm) if is_file else "-"

            result.append((info.tag_name, url, class_name, size))

        return result

    @property
    def plugins(self):
        result = []

        for info in self.result.assets.plugins:
            is_file = isinstance(info.esm, Path) and info.esm.is_file()
            class_name = str(self._path_exists_class(info.esm) if is_file else "")
            url = str(info.esm) if is_file else "not file"
            size = self._try_get_file_size_mb(info.esm) if is_file else "-"

            result.append((info.name, url, class_name, size))

        return result

    def _try_get_file_size_mb(self, path: Union[str, Path]):
        if isinstance(path, str):
            return None

        return f"{round(path.stat().st_size / 1024 / 1024, 3)} MB"

    def _path_exists_class(self, path: Union[str, Path]):
        if isinstance(path, str):
            return ""
        return "" if path.exists() else NO_EXISTS_PATH_CLASS


class ZeroDebugReporter:
    def generate_report_html(self, result: ZeroRenderResult) -> str:
        from instaui import zero

        return zero().to_html_str(lambda: _create_debug_report(ReportModel(result)))


NO_EXISTS_PATH_CLASS = "ex-no-exists-path"


def _create_debug_report(model: ReportModel):
    from instaui import ui, html

    ui.use_tailwind()
    ui.add_style(rf".{NO_EXISTS_PATH_CLASS} {{background-color: red;color: white;}}")
    box_style = "border-2 border-gray-200 p-4 place-center gap-x-4"

    with ui.column().classes("gap-2"):
        # base info
        with ui.grid(columns="auto 1fr").classes(box_style):
            html.span("file size:")
            html.span(f"{model.html_size_mb:.2f} MB")

        # import maps
        ui.heading("import maps")

        with ui.grid(columns="auto auto auto").classes(box_style):
            ui.text("name")
            ui.text("path or url")
            ui.text("size")

            ui.box(height="1px", width="100%", grid_column="1/-1").style(
                "border-top: 1px solid black;"
            )

            for name, url, size in model.import_maps_records:
                ui.text(name)
                ui.text(url)
                ui.text(size)

        # css links
        ui.heading("css links")

        with ui.grid(columns="1fr auto").classes(box_style):
            ui.text("path or url")
            ui.text("size")

            for link, link_class, size in model.css_links:
                html.span(link).classes(link_class)
                ui.text(size)

        # js links
        ui.heading("js links")
        with ui.column().classes(box_style):
            for link, link_class, size in model.js_links:
                html.span(link).classes(link_class)

        # custom components
        ui.heading("custom components")
        with ui.grid(columns="auto 1fr auto").classes(box_style):
            html.span("name")
            html.span("js file path")
            html.span("size")

            for name, url, class_name, size in model.vue_app_component:
                html.span(name)
                html.span(url).classes(class_name)
                ui.text(size)

        # plguins
        ui.heading("plugins")
        with ui.grid(columns="auto 1fr auto").classes(box_style):
            html.span("name")
            html.span("js file path")
            html.span("size")

            for name, url, class_name, size in model.plugins:
                html.span(name)
                html.span(url).classes(class_name)
                ui.text(size)



from __future__ import annotations
from typing import TYPE_CHECKING

from instaui.internal.ui.app_context import get_app


from .variable import Variable
from .container import Container, ScopeGuard
from .renderable import Renderable

if TYPE_CHECKING:
    from instaui.internal.ui.ref_base import RefBase
    from instaui.internal.ui.custom_var import CustomVar
    from instaui.internal.ui.const_data import ConstData
    from instaui.internal.ui.web_computed import WebComputed
    from instaui.internal.ui.js_computed import JsComputed

    from instaui.internal.ui.expr_computed import ExprComputed
    from instaui.internal.ui.web_watch import WebWatch
    from instaui.internal.ui.js_watch import JsWatch
    from instaui.internal.ui.expr_watch import ExprWatch

    from instaui.internal.ui.element_ref import ElementRef
    from instaui.internal.ui.web_event import WebEvent
    from instaui.internal.ui.js_event import JsEvent
    from instaui.internal.ui.expr_event import ExprEvent
    from instaui.internal.ui.imports import JSImport
    from instaui.internal.ui.js_fn import JsFn


class Scope(Container, Renderable):
    def __init__(self) -> None:
        super().__init__()

        self._id = get_app().next_scope_id()

        self._refs: list[RefBase] = []
        self._custom_refs: list[CustomVar] = []
        self._web_computeds: list[WebComputed] = []
        self._js_computeds: list[JsComputed] = []
        self._element_refs: list[ElementRef] = []
        self._expr_computeds: list[ExprComputed] = []
        self._run_method_records: list = []
        self._web_watchs: list[WebWatch] = []
        self._js_watchs: list[JsWatch] = []
        self._expr_watchs: list[ExprWatch] = []
        self._web_events: list[WebEvent] = []
        self._js_events: list[JsEvent] = []
        self._expr_events: list[ExprEvent] = []
        self._js_fns: list[JsFn] = []
        self._js_imports: list[JSImport] = []

        self.variable_order: list[Variable] = []
        self._renderables: list[Renderable] = []
        self._scope_guard = ScopeGuard()
        self._injected: set[Variable] = set()
        self._provided: set[Variable] = set()

    @property
    def id(self) -> int:
        return self._id

    def __enter__(self):
        app = get_app()
        app.push_scope(self)
        return super().__enter__()

    def __exit__(self, *_):
        get_app().pop_scope()
        return super().__exit__(*_)

    def inject(self, variable: Variable):
        self._injected.add(variable)

    def provide(self, variable: Variable):
        self._provided.add(variable)

    def register_import(self, js_import: JSImport):
        self._js_imports.append(js_import)

    def register_data(self, data: ConstData):
        self.variable_order.append(data)

    def register_custom_ref(self, ref: CustomVar):
        self._custom_refs.append(ref)
        self.variable_order.append(ref)

    def register_ref(self, ref: RefBase):
        self._refs.append(ref)
        self.variable_order.append(ref)

    def register_element_ref(self, ref: ElementRef):
        self._element_refs.append(ref)
        self.variable_order.append(ref)

    def register_js_computed(self, computed: JsComputed):
        self._js_computeds.append(computed)
        self.variable_order.append(computed)

    def register_expr_computed(self, computed: ExprComputed):
        self._expr_computeds.append(computed)
        self.variable_order.append(computed)

    def register_web_computed(self, web_computed: WebComputed):
        self._web_computeds.append(web_computed)
        self.variable_order.append(web_computed)

    def register_web_watch(self, web_watch: WebWatch):
        self._web_watchs.append(web_watch)

    def register_js_watch(self, js_watch: JsWatch):
        self._js_watchs.append(js_watch)

    def register_expr_watch(self, vue_watch: ExprWatch):
        self._expr_watchs.append(vue_watch)

    def register_web_event(self, event: WebEvent):
        self._web_events.append(event)
        self.variable_order.append(event)

    def register_js_event(self, event: JsEvent):
        self._js_events.append(event)
        self.variable_order.append(event)

    def register_expr_event(self, event: ExprEvent):
        self._expr_events.append(event)
        self.variable_order.append(event)

    def register_js_fn(self, fn: JsFn):
        self._js_fns.append(fn)
        self.variable_order.append(fn)

    def add_child(self, renderable: Renderable):
        self._renderables.append(renderable)

    def _bind_scope(self, scope: Scope):
        self._scope_guard.bind_scope(scope)

    def _release_scope(self):
        self._scope_guard.release_scope()



from __future__ import annotations
from typing import (
    Optional,
    Union,
)
from instaui.internal.ui.value_element import ValueElement
from instaui.internal.ui.element import Element
from instaui.internal.ui.event import EventMixin
from instaui.internal.ui.event_modifier import TEventModifier
from instaui.internal.ui.vfor import VFor
from instaui.internal.ui.disabled_element import DisabledElement

_T_Select_Value = Union[list[str], str]


class Select(DisabledElement, ValueElement[Union[list[str], str]]):
    def __init__(
        self,
        value: Union[_T_Select_Value, None] = None,
        *,
        model_value: Union[str, None] = None,
    ):
        """
        A dropdown select component for HTML-based UIs.

        Args:
            value (Union[_T_Select_Value, None]): The initial selected value(s).
            model_value (Union[str, None]): Reactive model binding for the selected value.
                                                          If provided, overrides the `value` prop.

        # Example:
        .. code-block:: python
            from instaui import html

            # Basic usage
            with html.select(value="foo"):
                html.option(text="foo", value="foo")
                html.option(text="bar", value="bar")

            # Using from_list for simple options
            html.select.from_list(["foo", "bar"], value="foo")

            # Using from_records for structured data
            html.select.from_records(
                [{"value": "1", "text": "One"}, {"value": "2", "text": "Two"}],
                value="1"
            )
        """

        super().__init__("select", value, is_html_component=True)

        if model_value is not None:
            self.props({"value": model_value})

    def on_change(
        self,
        handler: EventMixin,
        *,
        extends: Optional[list] = None,
        modifier: Optional[list[TEventModifier]] = None,
    ):
        """
        Registers an event handler for the 'change' event.

        Args:
            handler (EventMixin): The function or callable to execute when selection changes.

        """
        self.on("change", handler, extends=extends, modifier=modifier)
        return self

    @classmethod
    def from_list(
        cls,
        options: list,
        value: Union[_T_Select_Value, None] = None,
        *,
        model_value: Union[str, None] = None,
    ) -> Select:
        """
        Creates a Select from a simple list of string options.

        Args:
            options (list): List of strings used as both text and value for each option.
            value (Union[_T_Select_Value, None]): The initially selected value.
            model_value (Union[str, None]): Reactive model binding for the selected value.

        """

        with cls(value, model_value=model_value) as select:
            with VFor(options) as item:
                Select.Option(text=item, value=item)

        return select

    @classmethod
    def from_records(
        cls,
        options: list[dict],
        value: Union[_T_Select_Value, None] = None,
        *,
        model_value: Union[str, None] = None,
    ) -> Select:
        """
        Creates a Select from a list of dictionaries with 'text' and 'value' keys.

        Args:
            options (list[dict]): List of dicts containing 'text' (display) and 'value' (data) fields.
            value (Union[_T_Select_Value, None]): The initially selected value.
            model_value (Union[str, None]): Reactive model binding for the selected value.

        """
        with cls(value, model_value=model_value) as select:
            with VFor(options) as item:
                Select.Option(text=item["text"], value=item["value"])

        return select

    class Option(Element, DisabledElement):
        def __init__(
            self,
            *,
            text: Optional[str] = None,
            value: Optional[str] = None,
            disabled: Optional[bool] = None,
        ):
            """
            Represents an individual option within a Select component.

            Args:
                text (Optional[str]): Display text for the option.
                value (Optional[str]): Value associated with the option (submitted with form).
                disabled (Optional[bool]): Whether the option is disabled for selection.
            """
            super().__init__("option")

            self.props({"value": value, "disabled": disabled, "text": text})



from fastapi import APIRouter
from fastapi.responses import FileResponse
from instaui.internal.backend.fastapi.endpoints import ServerEndpoints
from . import asset_mapper


def create_router(router: APIRouter, endpoints: ServerEndpoints):
    @router.get(endpoints.FILE_ROUTER_URL)
    def _(hash_part: str, file_name: str) -> FileResponse:
        hash_part_with_extend_paths = hash_part.split("/", maxsplit=1)
        hash_part = hash_part_with_extend_paths[0]
        extend_path = (
            None
            if len(hash_part_with_extend_paths) == 1
            else hash_part_with_extend_paths[1]
        )

        folder = asset_mapper.get_folder_path(hash_part)
        if extend_path:
            folder = folder.joinpath(extend_path)
        local_file = folder / file_name

        return FileResponse(
            local_file, headers={"Cache-Control": "public, max-age=3600"}
        )



from pathlib import Path
from typing import Any, Callable, Optional, Union
from instaui.internal.runtime.zero.options import CdnResourceOption, ZeroOptions
from instaui.systems import file_path_system


class ZeroLauncher:
    def __init__(
        self,
        *,
        icons_svg_path: Optional[Union[str, Path, Callable[[], Path]]] = None,
        cdn_resource_overrides: Optional[
            Union[list[CdnResourceOption], CdnResourceOption]
        ] = None,
        debug: bool = False,
    ):
        cdns = (
            cdn_resource_overrides
            if isinstance(cdn_resource_overrides, list)
            else [cdn_resource_overrides]
            if cdn_resource_overrides
            else None
        )

        icons_svg_path = (
            file_path_system.resolve_relative_path(icons_svg_path)
            if isinstance(icons_svg_path, str)
            else icons_svg_path
        )
        self._options = ZeroOptions(debug, icons_svg_path, cdns)

    def to_html(self, render_fn: Callable[..., Any], *, file: Union[str, Path]):
        """
        Generates a static HTML file by rendering the provided function.

        Args:
            render_fn (Callable[..., Any]): A callable that defines the content to be rendered into HTML.
                                        This function typically contains UI component calls.
            file (Union[str, Path]): The output path where the generated HTML file will be saved.
                                    Can be a string or a Path object.

        # Example:
        .. code-block:: python
            from instaui import ui, zero

            def page():
                ui.text("Hello, World!")

            zero().to_html(page, file="output.html")
        """
        file = file_path_system.resolve_relative_path(file)
        file.write_text(self.to_html_str(render_fn), "utf8")

    def to_html_str(self, render_fn: Callable[..., Any]):
        """
        Generates a static HTML file content string by rendering the provided function.

        Args:
            render_fn (Callable[..., Any]): A function that defines the UI content by calling UI components.
                                            It is executed to capture the view structure for static rendering.

        # Example:
        .. code-block:: python
            from instaui import ui, zero

            def page():
                ui.text("Hello, World!")

            html_str = zero().to_html_str(page)
        """
        from instaui.internal.runtime.zero.zero_runtime import ZeroRuntime

        return ZeroRuntime(self._options).run(render_fn)

    def to_debug_report(self, render_fn: Callable[..., Any], *, file: Union[str, Path]):
        """
        Generates a debug report for the static HTML output, including file size and resource usage analysis.

        Args:
            render_fn (Callable[..., Any]): A function that defines the UI content by calling UI components.
                                            It is used to generate the HTML for analysis.
            file (Union[str, Path]): The output file path where the debug report will be saved.

        # Example:
        .. code-block:: python
            from instaui import ui, zero

            def page():
                ui.text("Hello, World!")

            zero().to_debug_report(page, file="debug_report.html")
        """
        file = file_path_system.resolve_relative_path(file)

        # render_result = ZeroRuntime().render_html(render_fn, self._options)

        # file.write_text(
        #     ZeroDebugReporter().generate_report_html(render_result), encoding="utf8"
        # )



from itertools import chain, groupby
from instaui.internal.assets.base import AssetsDeclaration
from instaui.internal.assets.component_dep import ComponentDependencyRegistry
from instaui.internal.assets.css_assets import CssAssetCollection
from instaui.internal.assets.snapshot import AssetsSnapshot, ComponentDependencySnapshot
from instaui.internal.assets.style_assets import StyleTag
from .processors.base import AssetsProcessor


class AssetsPipeline:
    def __init__(self, processors: list[AssetsProcessor]):
        self._processors = processors

    def run(
        self, global_assets: AssetsDeclaration, local_assets: AssetsDeclaration
    ) -> AssetsSnapshot:
        merged = self.merge(global_assets, local_assets)

        for processor in self._processors:
            processor.process(merged)

        return AssetsSnapshot(
            js_links=list(merged.js_links),
            css_links=list(merged.css_links.iter_css_links()),
            style_tags=merged.style_tags,
            script_tags=merged.script_tags,
            import_maps=merged.import_maps,
            favicon=merged.favicon,
            plugins=merged.plugins,
            component_dependencies=_convert_component_snapshot(
                merged.component_dependencies
            ),
        )

    def merge(
        self, global_assets: AssetsDeclaration, local_assets: AssetsDeclaration
    ) -> AssetsDeclaration:
        assert local_assets is not None, "should be called inside a page() context"

        plugins = global_assets.plugins | local_assets.plugins

        component_dependencies = ComponentDependencyRegistry()

        for record in chain(
            global_assets.component_dependencies.records,
            local_assets.component_dependencies.records,
        ):
            component_dependencies.add(record.component, record.dependency)

        js_links = global_assets.js_links | local_assets.js_links
        css_links = CssAssetCollection.merge(
            global_assets.css_links, local_assets.css_links
        )
        style_tags = _merged_style_tags(
            global_assets.style_tags, local_assets.style_tags
        )
        script_tags = global_assets.script_tags + local_assets.script_tags
        import_maps = {**global_assets.import_maps, **local_assets.import_maps}
        favicon = global_assets.favicon or (
            local_assets.favicon if local_assets else None
        )

        component_extensions = global_assets.component_extensions.overridden_by(
            local_assets.component_extensions
        )

        return AssetsDeclaration(
            js_links=js_links,
            css_links=css_links,
            style_tags=style_tags,
            script_tags=script_tags,
            import_maps=import_maps,
            favicon=favicon,
            plugins=plugins,
            component_dependencies=component_dependencies,
            component_extensions=component_extensions,
        )


def _merged_style_tags(global_tags: list[StyleTag], local_tags: list[StyleTag]):
    tags = global_tags + local_tags
    sorted_tags = sorted(tags, key=lambda tag: tag.group_id or "")

    return [
        StyleTag("\n".join(tag.content for tag in group_tags), group_id=group_id)
        for group_id, group_tags in groupby(sorted_tags, lambda tag: tag.group_id)
    ]


def _convert_component_snapshot(
    registry: ComponentDependencyRegistry,
) -> list[ComponentDependencySnapshot]:
    return [
        ComponentDependencySnapshot(
            tag=record.dependency.tag_name,
            esm=record.dependency.esm,
        )
        for record in registry.records
    ]
